# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2019


"""
REST API bindings for IBM® Streams Cloud Pak for Data build service.

**********************
Streams Build REST API
**********************

The REST Build API provides programmatic support for creating, submitting and managing Streams builds. 
You can use the REST Build API from any application that can establish an HTTPS connection to the server 
that is running the build service.  The current support includes methods for managing toolkits in the build service 
and for retrieving base images for Edge image builds.

Cloud Pak for Data
==================

:py:meth:`~streamsx.build.BuildService.of_endpoint` or :py:meth:`~streamsx.build.BuildService.of_service` 
is the entry point to using the Streams Build REST API bindings, returning a :py:class:`~streamsx.build.BuildService`.


.. seealso:: :ref:`sas-main`
"""

__all__ = ['BuildService']

import os
import time
import json
import logging
import re
import requests
import socket
import tempfile
import warnings
from pprint import pformat
from urllib import parse
from zipfile import ZipFile

import streamsx.topology.context

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__

from streamsx import st
import streamsx.rest
import streamsx.rest_primitives

logger = logging.getLogger('streamsx.build')


class BuildService(streamsx.rest._AbstractStreamsConnection):
    """IBM Streams build service.

    An instance of a `BuildService` is created using :py:meth:`of_endpoint` or :py:meth:`of_service`.

    .. versionadded:: 1.13
    """
    def __init__(self, username=None, password=None, resource_url=None, buildpools_url=None, auth=None):
        """specify username, password, and resource_url"""
        streamsx._streams._version._mismatch_check(__name__)
        if auth:
            pass
        elif username and password:
            pass
        else:
            raise ValueError("Must supply either an authentication token or a username, password"
                             " to the BuildService constructor.")

        if not resource_url and 'STREAMS_REST_URL' in os.environ:
            resource_url = os.environ['STREAMS_REST_URL']
        
        self._build_url = resource_url
        self._buildpools_url = buildpools_url
        if auth:
            self.rest_client = streamsx.rest_primitives._StreamsRestClient(auth)
        else:
            self.rest_client = streamsx.rest_primitives._StreamsRestClient._of_basic(username, password)
        self.rest_client._sc = self
        self.session = self.rest_client.session

    @property
    def resource_url(self):
        """str: Endpoint URL for IBM Streams REST build API.

        """
        if self._build_url:
            return re.sub('/builds$','/resources', self._build_url)
        return None

    def get_resources(self):
        """Retrieves a list of all known Streams high-level Build REST resources.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives.RestResource`: List of all Streams high-level Build REST resources.

        """
        return super().get_resources()

    def get_toolkits(self, name=None):
        """Retrieves a list of all installed Streams Toolkits.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives.Toolkit`: List of all Toolkit resources.

        Args:
           name(str): Return toolkits matching name as a regular expression.

        """
        return self._get_elements('toolkits', streamsx.rest_primitives.Toolkit, name=name)
     
    def get_toolkit(self, id):
        """Retrieves available toolkit matching a specific toolkit ID.

        Args:
            id (str): Toolkit identifier to retrieve.  This is the name and 
                      version of a toolkit.  For sample, `com.ibm.streamsx.rabbitmq-1.1.3`

        Returns:
            :py:class:`~.rest_primitives.Toolkit`: Toolkit matching `id`.

        Raises:
            ValueError: No matching toolkit exists.

        """
        return self._get_element_by_id('toolkits', streamsx.rest_primitives.Toolkit, id)

    def upload_toolkit(self, path):
        """
        Upload a toolkit from a directory in the local filesystem to 
        the Streams build service.

        Multiple versions of a toolkit may be uploaded as long as each has
        a unique version.  If a toolkit is uploaded with a name and version
        matching an existing toolkit, it will not replace the existing
        toolkit, and ``None`` will be returned.
       
        Args:
            path(str): The path to the toolkit directory in the local filesystem.
        Returns:
            Toolkit: The created Toolkit, or ``None`` if it was not uploaded.

        """
        # Handle path does not exist, is not readable, is not a directory
        if not os.path.isdir(path):
            raise ValueError('"' + path + '" is not a path or is not readable')

        # Create a named temporary file
        with tempfile.NamedTemporaryFile(suffix='.zip') as tmpfile:
            filename = tmpfile.name
        
            basedir = os.path.abspath(os.path.join(path, os.pardir))

            with ZipFile(filename, 'w') as zipfile:
                for root, dirs, files in os.walk(path):
                    # Write the directory entry
                    relpath = os.path.relpath(root, basedir)
                    zipfile.write(root, relpath)
                    for file in files:
                        zipfile.write (os.path.join(root, file), os.path.join(relpath, file))
                zipfile.close()
            
                with open(filename, 'rb') as toolkit_fp:
                    res = self.rest_client.session.post(streamsx.rest_primitives.Toolkit._toolkits_url(self),
                        headers = {'Accept' : 'application/json',
                                   'Content-Type' : 'application/zip'},
                        data=toolkit_fp,
                        verify=self.rest_client.session.verify)
                    streamsx.rest_primitives._handle_http_errors(res)
                    new_toolkits = list(streamsx.rest_primitives.Toolkit(t, self.rest_client) for t in res.json()['toolkits'])

                    # It may be possible to upload multiple toolkits in one 
                    # post, but we are only uploading a single toolkit, so the
                    # list of new toolkits is expected to contain only one 
                    # element, and we return it.  It is also possible that no 
                    # new toolkit was returned.

                    if len(new_toolkits) >= 1:
                        return new_toolkits[0]    
                    return None

    def _get_buildPools(self, name=None):
        """Returns the build pools via the new resource type 'buildPools' from Edge enabled build services.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives._BuildPool`: A list of _BuildPool instances or ``None`` if build pools are not available.
        """
        buildpools = self._get_elements('buildPools', streamsx.rest_primitives._BuildPool, name=name)
        if buildpools is None:
            # workaround as long as 'buildPools' resource is not available
            buildpools_url = self._buildpools_url
            if not buildpools_url:
                # We do NOT get the deprecated BuildPool REST resource here as
                # it will not be a build pool of 'image' type
                return None
            buildpools_json = self.rest_client.make_request(buildpools_url)['buildPools']
            buildpools = [streamsx.rest_primitives._BuildPool(json_rep, self.rest_client) for json_rep in buildpools_json]

        return buildpools
    
    def get_base_images(self):
        """Retrieves a list of all installed base images for Edge applications.

        Returns:
            :py:obj:`list` of :py:class:`~.rest_primitives.BaseImage`: List of all base images, ``None`` if there are no base images

        .. versionadded:: 1.15
        """
        buildpools = self._get_buildPools()
        if not buildpools:
            # None or empty list
            return None
        images = []
        for pool in buildpools:
            if hasattr(pool, 'baseimages'):
                images.extend([streamsx.rest_primitives.BaseImage(json_rep, self.rest_client) for json_rep in self.rest_client.make_request(pool.baseimages)['images']])
        if images:
            return images
        return None

    @staticmethod
    def _is_service_def(config):
        return 'connection_info' in config and config.get('type', None) == 'streams' and 'service_token' in config

    @staticmethod
    def _find_service_def(config):
        if BuildService._is_service_def(config):
            service = config
        else:
            service = config.get(streamsx.topology.context.ConfigParams.SERVICE_DEFINITION)

        if service and BuildService._is_service_def(service):
            svc_info = {}
            svc_info['connection_info'] = service['connection_info']
            svc_info['type'] = service['type']
            svc_name = service['connection_info']['serviceRestEndpoint'].split('/')[-1]
            svc_info['service_name'] = svc_name
            try:
                # Get a new token as we don't know how much
                # time has passed since the cell containing
                # get_service_instance_details was run
                from icpd_core import icpd_util
                svc_info['service_token'] = icpd_util.get_instance_token(name=svc_name)
                svc_info['service_token_expire'] = int((time.time() + 19 * 60)*1000)
            except:
                svc_info['service_token'] = service['service_token']
                svc_info['service_token_expire'] = int((time.time() + 5 * 60)*1000)
            if 'user_token' in service:
                svc_info['user_token'] = service['user_token']
            return svc_info
        return None

    @staticmethod
    def of_service(config):
        """Connect to a Cloud Pak for Data IBM Streams build service instance.

        The instance is specified in `config`. The configuration may be code injected from the list of services
        in a Jupyter notebook running in ICPD or manually created. The code that selects a service instance by name is::

            from icpd_core import ipcd_util
            cfg = icpd_util.get_service_details(name='instanceName', instance_type='streams')

            buildService = BuildService.of_service(cfg)
            
        SSL host verification is disabled by setting :py:const:`~streamsx.topology.context.ConfigParams.SSL_VERIFY`
        to ``False`` within `config` before calling this method::

            from icpd_core import ipcd_util
            cfg = icpd_util.get_service_details(name='instanceName', instance_type='streams')

            cfg[ConfigParams.SSL_VERIFY] = False
            buildService = BuildService.of_service(cfg)

        Args:
            config(dict): Configuration of IBM Streams service instance.

        Returns:
            :py:class:`BuildService`: Connection to Streams build service.

        .. note:: Only supported when running within the ICPD cluster,
            for example in a Jupyter notebook within an ICPD project.

        .. versionadded:: 1.15
        """
        service = BuildService._find_service_def(config)
        if not service:
            raise ValueError()
        # service_name is the instance name
        service_name = service['service_name']
        auth = streamsx.rest_primitives._ICPDAuthHandler(service_name, service['service_token'])
        build_url = BuildService._root_from_endpoint(service['connection_info'].get('serviceBuildEndpoint'))
        buildpools_url = service['connection_info'].get('serviceBuildPoolsEndpoint', None)
        sc = BuildService(resource_url=build_url, buildpools_url=buildpools_url, auth=auth)
        if streamsx.topology.context.ConfigParams.SSL_VERIFY in config:
            sc.rest_client.session.verify = config[streamsx.topology.context.ConfigParams.SSL_VERIFY]
        return sc

    @staticmethod
    def of_endpoint(endpoint=None, service_name=None, username=None, password=None, verify=None):
        """
        Connect to a Cloud Pak for Data IBM Streams build service instance.

        Two configurations are supported.

        .. rubric:: Integrated configuration

        The build service is bound to a Streams instance and is defined
        using the Cloud Pak for Data deployment endpoint (URL) and
        the Streams service name.

        The endpoint is passed in as `endpoint` defaulting the the
        environment variable ``CP4D_URL``.
        An example is `https://cp4d_server:31843`.

        The Streams service name is passed in as `service_name` defaulting
        to the environment variable ``STREAMS_INSTANCE_ID``.

        .. rubric:: Standalone configuration

        A build service is independent of a Streams instance and is defined
        using the build service endpoint.

        The endpoint is passed in as `endpoint` defaulting the the
        environment variable ``STREAMS_BUILD_URL``.
        An example is `https://build_service:34679`.

        No service name is specified thus `service_name` should be passed
        as ``None`` or not set.

        Args:
            endpoint(str): Endpoint defining the build service.
            service_name(str): Streams instance name for a integrated configuration.  This value is ignored for a standalone configuration.
            username(str): User name to authenticate as. Defaults to the environment variable ``STREAMS_USERNAME`` or the operating system identifier if not set.
            password(str): Password for authentication. Defaults to the environment variable ``STREAMS_PASSWORD`` or the operating system identifier if not set.
            verify: SSL verification. Set to ``False`` to disable SSL verification. Defaults to SSL verification being enabled.
       
        Returns:
            :py:class:`BuildService`: Connection to Streams build service or ``None`` if insufficient configuration was provided.
        """
        possible_integ = True
        if not endpoint:
            endpoint = os.environ.get('CP4D_URL')
            if not endpoint:
                endpoint = os.environ.get('STREAMS_BUILD_URL')
                if not endpoint:
                    return None
                possible_integ = False
                if service_name:
                    warnings.warn("Service name ignored for standalone configuration", UserWarning, stacklevel=2)
                    service_name = None

        if possible_integ and not service_name:
            service_name = os.environ.get('STREAMS_INSTANCE_ID')
        if not password:
            password = os.environ.get('STREAMS_PASSWORD')
        if not password:
            return None
        username = streamsx.rest_primitives._get_username(username)

        if service_name:
            # this is an integrated config
            auth=streamsx.rest_primitives._ICPDExternalAuthHandler(endpoint, username, password, verify, service_name)
            build_url = BuildService._root_from_endpoint(auth._cfg['connection_info'].get('serviceBuildEndpoint'))
            buildpools_ep = auth._cfg['connection_info'].get('serviceBuildPoolsEndpoint', None)
            if buildpools_ep:
                buildpools_url = buildpools_ep
            else:
                buildpools_url = None
            service_name=auth._cfg['service_name']
            sc = BuildService(resource_url=build_url, buildpools_url=buildpools_url, auth=auth)
            if verify is not None:
                sc.rest_client.session.verify = verify

        else:
            # This is a stand-alone config
            parsed = parse.urlparse(endpoint)
            build_url = parse.urlunparse((parsed.scheme, parsed.netloc, "/streams/rest/resources", None, None, None))

            # Create a connection using basic authentication.  This will be
            # used to attempt to discover the security service.  If there
            # is no security service, this connection will continue to be
            # used; otherwise, it will be replaced by one using the 
            # security service.
            sc = BuildService(resource_url=build_url, username=username, password=password)
            if verify is not None:
                sc.rest_client.session.verify = verify
            for resource in sc.get_resources():
                if resource.name == 'accessTokens':
                    atinfo = parse.urlparse(resource.resource)
                    try:
                        # If we are external to the cluster then
                        # We can't resolve the security manager
                        # so revert to basic auth
                        socket.gethostbyname(atinfo.hostname)
                        auth=streamsx.rest_primitives._JWTAuthHandler(resource.resource, username, password, verify)
                        sc = BuildService(resource_url=build_url, buildpools_url=None, auth=auth)
                        if verify is not None:
                            sc.rest_client.session.verify = verify
                        break
                    except:
                        pass
            else:
                # No security service could be found.  We can try to proceed 
                # with basic authentication.
                pass
 
        return sc

    @staticmethod
    def _root_from_endpoint(endpoint):
        import urllib.parse as up
        esu = up.urlsplit(endpoint)
        # CPD 2.5
        if esu.path.startswith('/streams-build/instances/'):
            return endpoint.replace('/streams-build/instances', '/streams-build-resource/instances', 1)

        if not esu.path.startswith('/streams/rest/builds'):
            return None

        es = endpoint.split('/')
        root_url = endpoint.split('/streams/rest/builds')[0]
        resource_url = root_url + '/streams/rest/resources'
        return resource_url

    def __str__(self):
        return pformat(self.__dict__)




# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2019

import sys
import sysconfig
import os
import argparse
import urllib3
import shutil
import streamsx.rest
import pkg_resources
from streamsx.spl.op import _main_composite
from streamsx.spl.toolkit import add_toolkit
from streamsx.topology.context import submit, ConfigParams
from streamsx.build import BuildService
import xml.etree.ElementTree as ET
from glob import glob
import re
from io import StringIO


_FACADE=['--prefer-facade-tuples', '-k']

def main(args=None):
    """ Mimics SPL compiler sc against the ICP4D build service.
    """
    cmd_args = _parse_args(args)
    topo = _create_topo(cmd_args)

    # Add main composite toolkit
    _add_toolkits(cmd_args, topo)

    # Get dependencies for app, if any at all
    dependencies = _parse_dependencies()
    # if dependencies and if -t arg, find & add local toolkits
    spl_path = cmd_args.spl_path
    if spl_path is None:
        spl_path = os.environ.get('STREAMS_SPLPATH')
    if dependencies and spl_path:
        tool_kits = spl_path.split(':')
        # Check if any dependencies are in the passed in toolkits, if so add them
        _add_local_toolkits(tool_kits, dependencies, topo, verify_arg = False if cmd_args.disable_ssl_verify else None)

    sr = _submit_build(cmd_args, topo)
    return _move_bundle(cmd_args, sr)

def _parse_dependencies():
    # Parse info.xml for the dependencies of the app you want to build sab file for
    deps = {} # name - version
    if os.path.isfile('info.xml'):
        # Open XML file and strip all namespaces from XML tags
        # Ex. <info:dependencies> -> <dependencies>
        XML_data = None
        with open('info.xml', 'r') as file:
            XML_data = file.read()
        it = ET.iterparse(StringIO(XML_data))
        for _, el in it:
            if '}' in el.tag:
                el.tag = el.tag.split('}', 1)[1]
        root = it.root

        # Get dependency tag
        dependency_elements = root.findall('dependencies')
        assert len(dependency_elements) == 1, 'Should only contain 1 dependency tag'

        # Get toolkits tags
        dependencies = dependency_elements[0].findall('toolkit')
        for toolkit in dependencies:
            name = toolkit.find('name').text
            version = toolkit.find('version').text
            deps[name] = version
    return deps

def _add_local_toolkits(toolkit_paths, dependencies, topo, verify_arg):
    """ A helper function that given the local toolkit paths, and the dependencies (ie toolkits that the apps depends on),
    finds the latest dependency w/ correct version (could be local or on the buildserver) and checks whether it exists locally, and if so adds it.
    If the latest dependency w/ correct version is located on the buildserver, don't need to do anything

    Arguments:
        toolkit_paths {List} -- A list of local toolkit directories, each could be a toolkit directory (has info.xml directly inside),
                or a directory consisting of toolkits (no info.xml directly inside)
        dependencies {Dictionary} -- A dictionary consisting of elements where the key is the name of the dependency, and the value is its version.
        topo {topology Object} -- [description]
        verify_arg {type} -- Disable SSL verification.
    """
    build_server = BuildService.of_endpoint(verify=verify_arg)

    local_toolkits = _get_all_local_toolkits(toolkit_paths)
    remote_toolkits = build_server.get_toolkits()

    all_toolkits = local_toolkits + remote_toolkits

    # Iterate through all dependencies (toolkits) needed to build sab file, and check if the dependency already exist on the build server
    for dependency_name, dependency_version in dependencies.items():

        # All the toolkits that match dependency_name regardless of version
        matching_toolkits = [toolkit for toolkit in all_toolkits if toolkit.name == dependency_name]

        if not matching_toolkits:
            # No toolkits exists on the remote buildserver or locally that can satsify this dependency, continue and let SPL compiler throw error
            continue

        # Toolkit with the highest verion that still satisfies the dependency_version
        latest_compatible_toolkit = None

        # Go through all toolkits, find the highest/latest version, given by latest_compatible_toolkit
        for tk in matching_toolkits:
            # Check if current toolkit tk satisfies the version requirement
            if _check_correct_version(tk, dependency_version):
                # if we have a toolkit that already satisfies the version requirement, and tk also satisfies its
                # compare tk against the highest version we have seen thus far
                if latest_compatible_toolkit:
                    latest_version_so_far = pkg_resources.parse_version(latest_compatible_toolkit.version)
                    curr_tk_version = pkg_resources.parse_version(tk.version)

                    # if it is a later version, update our toolkit
                    if curr_tk_version > latest_version_so_far:
                        latest_compatible_toolkit = tk
                else:
                    latest_compatible_toolkit = tk

        # Check if latest_compatible_toolkit is local, bc we then need to add it
        if isinstance(latest_compatible_toolkit, _LocalToolkit):
            # Latest toolkit is local, upload it
            add_toolkit(topo, latest_compatible_toolkit.path)
            # print("Latest dependency {} with version {} is not on buildserver, adding it".format(latest_compatible_toolkit.name, latest_compatible_toolkit.version))
        # latest_compatible_toolkit exists on buildserver, thus don't need to do anything

def _get_all_local_toolkits(toolkit_paths):
    """ A helper function that given the local toolkit paths, creates and returns a list of local toolkit objects

    Arguments:
        toolkit_paths {List} -- A list of local toolkit directories, each could be a toolkit directory (has info.xml directly inside),
            or a directory consisting of toolkits (no info.xml directly inside)

    Returns:
        local_toolkits {List} -- list of local toolkit objects
    """

    local_toolkits_paths = [] # toolkit paths
    local_toolkits = [] # toolkit objects

    #  For each path, check if its a SPL toolkit (has info.xml directly inside)
    #  or a directory consisting of toolkits (no info.xml directly inside)
    for x in toolkit_paths:
        path = os.path.join(x, '') # Adds a trailing '/' if not already present
        if _is_local_toolkit(path):
            local_toolkits_paths.append(path)
        else:
            # directory consisting of toolkits, get list and check which are local_toolkits, add them
            sub_directories = glob(path + "*/") # All values in sub_directories will have a trailing '/'
            for y in sub_directories:
                if _is_local_toolkit(y):
                    local_toolkits_paths.append(y)

    # For each direct path to a toolkit, parse the name & version, and convert it to a _LocalToolkit object
    for tk_path in local_toolkits_paths:
        if os.path.isfile(tk_path + 'info.xml'):
            # Get the name & version of the toolkit that is in the directory tk_path
            root = ET.parse(tk_path + 'info.xml').getroot()
            identity = root.find('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}identity')
            name = identity.find('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}name')
            version = identity.find('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}version')

            toolkit_name = name.text
            toolkit_version = version.text

            tk = _LocalToolkit(toolkit_name, toolkit_version, tk_path)
            local_toolkits.append(tk)

    return local_toolkits

def _is_local_toolkit(dir):
    # Checks if dir is a toolkit (contains toolkit.xml or info.xml)
    for fn in ['toolkit.xml', 'info.xml']:
        if os.path.isfile(os.path.join(dir, fn)):
            return True
    return False

class _LocalToolkit:
    def __init__(self, name, version, path):
        self.name = name
        self.version = version
        self.path = path

def _check_correct_version(toolkit, dependency_range):
    """ Checks if toolkit's version satisfies the dependency_range, if yes return true, else false

    Arguments:
        toolkit {_LocalToolkit} -- A _LocalToolkit object, that we need to check if its version is equal to or is contained in dependency_range
        dependency_range {String} -- A string of the form '1.2.3' or [3.0.0,4.0.0)' that represents a version or range of versions that is acceptable
    """
    # Convert it to version
    toolkit_ver = pkg_resources.parse_version(toolkit.version)

    # Check if dependency_range is a single # or a range (single # won't contain brackets or parenthesis)
    temp = ['(', ')', '[', ']']
    if not any(x in dependency_range for x in temp):
        required_version = pkg_resources.parse_version(dependency_range)
        if required_version == toolkit_ver:
            return True
        return False

    # Dependency_range is confirmed to be a range, check if left & right bound is inclusive or exclusive
    left_inclusive = None
    right_inclusive = None

    satisfies_left = False
    satisfies_right = False

    # Check left bound
    if dependency_range[0] == '[':
        left_inclusive = True
    else:
        left_inclusive = False

    # Check right bound
    if dependency_range[-1] == ']':
        right_inclusive = True
    else:
        right_inclusive = False

    # Remove parenthesis and brackets from range and split by ',' to get lower and upper bounds, then convert to version
    # Version also handles the case '[4.5.6,5.2)' where 5.2 > 4.5.6
    bounds = re.sub('[()\[\]]', '', dependency_range).split(',')
    left_bound = pkg_resources.parse_version(bounds[0])
    right_bound = pkg_resources.parse_version(bounds[1])

    # Check that toolkit_ver satisfies its left and right bounds
    # if left is '[' check that version satisfies it
    if left_inclusive:
        if toolkit_ver >= left_bound:
            satisfies_left = True
    else: # left is '('
        if toolkit_ver > left_bound:
            satisfies_left = True
    # if right is ']' check that version satisfies it
    if right_inclusive:
        if toolkit_ver <= right_bound:
            satisfies_right = True
    else: # right is ')'
        if toolkit_ver < right_bound:
            satisfies_right = True

    # If toolkit.version satisfies left and right bound, then its a valid toolkit dependency, return True
    if satisfies_left and satisfies_right:
        return True
    return False

def _submit_build(cmd_args, topo):
     cfg = {}
     # Only supported for remote builds.
     cfg[ConfigParams.FORCE_REMOTE_BUILD] = True
     if cmd_args.disable_ssl_verify:
         cfg[ConfigParams.SSL_VERIFY] = False
         urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
     _sc_options(cmd_args, cfg)
     return submit('BUNDLE', topo, cfg)

def _move_bundle(cmd_args, sr):
    out_dir = cmd_args.output_directory
    if not out_dir:
        out_dir = 'output'
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)
    if 'jobConfigPath' in sr:
        jco = sr['jobConfigPath']
        shutil.move(jco, os.path.join(out_dir, os.path.basename(jco)))
    if not 'bundlePath' in sr:
        return 1
    sab = sr['bundlePath']
    shutil.move(sab, os.path.join(out_dir, os.path.basename(sab)))
    return 0

def _sc_options(cmd_args, cfg):
    args = []
    if cmd_args.prefer_facade_tuples:
        args.append(_FACADE[0])

    if cmd_args.ld_flags:
        args.append('--ld-flags=' + str(cmd_args.ld_flags))
    if cmd_args.cxx_flags:
        args.append('--cxx-flags=' + str(cmd_args.cxx_flags))
    if cmd_args.cppstd:
        args.append('--c++std=' + str(cmd_args.cppstd))
    if cmd_args.data_directory:
        args.append('--data-directory=' + str(cmd_args.data_directory))
    if cmd_args.compile_time_args: # sc -M my::App hello=a,b,c foo=bar -> compile_time_args = ['hello=a,b,c', 'foo=bar']
        # Check if '=' is NOT present in the compile_time_args, this implies that we have a .splmm file
        if any('=' not in arg for arg in cmd_args.compile_time_args):
            # If we have both SPLMM args and SPL CompileTimeValue args, then main composite is a .splmm file
            # _SPLMM_OPTIONS Should contain both regular SPL compile_time_args and SPLMM args to preserve ordering
            cfg[ConfigParams._SPLMM_OPTIONS] = cmd_args.compile_time_args
            # Just the SPL CompileTimeValue args
            args.extend([x for x in cmd_args.compile_time_args if '=' in x])
        else:
            # No SPLMM args, so compiling a regular .spl file
            args.extend(cmd_args.compile_time_args)
    if args:
        cfg[ConfigParams.SC_OPTIONS] = args

def _add_toolkits(cmd_args, topo):
    _add_cwd(cmd_args, topo)

def _add_cwd(cmd_args, topo):
    cwd = os.getcwd()
    if _is_likely_toolkit(cwd):
        add_toolkit(topo, cwd)

def _is_likely_toolkit(tkdir):
    """
    Determine if a directory looks like an SPL toolkit.
    """
    if _is_local_toolkit(tkdir):
        return True
    for dirpath, dirnames, filenames in os.walk(tkdir):
        for fn in filenames:
            if fn.endswith('.spl'):
                return True
            if fn.endswith('.splmm'):
                return True
            if fn.endswith('._cpp.cgt'):
                return True
        if '.namespace' in filenames:
            return True
        if 'native.function' in dirnames:
            return True

def _create_topo(cmd_args):
    # the private function permits main composites w/o a namespace
    topo,invoke = _main_composite(kind=cmd_args.main_composite)
    return topo

def _parse_args(args):
    """ Argument parsing
    """
    cmd_parser = argparse.ArgumentParser(description='SPL compiler (sc) alias for build service.')
    cmd_parser.add_argument('--main-composite', '-M', required=True, help='SPL Main composite', metavar='name')

    cmd_parser.add_argument('--spl-path', '-t', help='Set the toolkit lookup paths. Separate multiple paths with :. Each path is a toolkit directory, a directory of toolkit directories, or a toolkitList XML file. This path overrides the STREAMS_SPLPATH environment variable.')
    cmd_parser.add_argument('--optimized-code-generation', '-a', action='store_true', help='Generate optimized code with less runtime error checking.')
    cmd_parser.add_argument('--no-optimized-code-generation', action='store_true', help='Generate non-optimized code with more runtime error checking. Do not use with the --optimized-code-generation option.')
    cmd_parser.add_argument(*_FACADE, action='store_true', help='Generate the facade tuples when it is possible.')

    cmd_parser.add_argument('compile_time_args', help='arguments that are passed in at compile time.', nargs='*', metavar='compile-time-args')
    cmd_parser.add_argument('--ld-flags', '-w', help='Pass the specified flags to ld while linking occurs.')
    cmd_parser.add_argument('--cxx-flags', '-x', help='Pass the specified flags to the C++ compiler during the build.')
    cmd_parser.add_argument('--c++std', help='Specify the language level for the underlying C++ compiles.', dest='cppstd')
    cmd_parser.add_argument('--data-directory', help='Specifies the location of the data directory to use.')
    cmd_parser.add_argument('--output-directory', help='Specifies a directory where the application artifacts are placed.')

    _buildservice_args(cmd_parser)
    _deprecated_args(cmd_parser)

    return cmd_parser.parse_args(args)

def _buildservice_args(cmd_parser):
    rem = cmd_parser.add_argument_group('build service arguments', 'Arguments specific to use of the build service. Not supported by sc.')

    cmd_parser.add_argument('--disable-ssl-verify', action='store_true', help='Disable SSL verification.')

def _deprecated_args(cmd_parser):
    dep = cmd_parser.add_argument_group('deprecated arguments', 'Arguments that have been deprecated by sc and are ignored')

    dep.add_argument('--static-link', '-s', action='store_true')
    dep.add_argument('--standalone-application', '-T', action='store_true')
    dep.add_argument('--set-relax-fusion-relocatability-restartability', '-O', action='store_true')

    dep.add_argument('--checkpoint-directory', '-K', action='store', metavar='path')
    dep.add_argument('--profiling-sampling', '-S', action='store', metavar='rate')

if __name__ == '__main__':
    rc = main()
    sys.exit(rc)



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2020

"""
Definition of common types used for type hints.
"""

from typing import Any, Type, Union, NamedTuple

Schema = Union[
    'streamsx.topology.schema.StreamSchema',
    'streamsx.topology.schema.CommonSchema',
    str, NamedTuple]



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2019

import collections
import inspect
import typing
import sys
import streamsx.topology.schema

_SchemaHints = collections.namedtuple('_SchemaHints', ['type_', 'schema'])

def schema_iterable(fn, topo):
    try:
        return _schema_iterable(fn, topo)
    except TypeError:
        raise
    except:
        return

def _schema_iterable(fn, topo):
    if inspect.isroutine(fn):
        _check_arg_count(fn, 0)
    elif callable(fn):
        fn = fn.__call__
        _check_arg_count(fn, 0)
    elif hasattr(fn, '__iter__'):
        fn = fn.__iter__
    else:
        return

    if not topo.type_checking:
        return

    type_ = _fn_return_hint(fn)
    if type_ is not None:
        if hasattr(type_, '__origin__') and hasattr(type_, '__args__'):
            if len(type_.__args__) == 1:
                et = type_.__args__[0]
                if typing.Iterable[et] == type_:
                    return _schema_from_type(et)
                if typing.Iterator[et] == type_:
                    return _schema_from_type(et)

def _fn_return_hint(fn):
    hints = typing.get_type_hints(fn)
    if hints and 'return' in hints:
        return hints['return']
        
# Check the callable for a filter
#
# - Can be passed a single argument
# - Has an argument type that is compatible the the schema type.
# - TODO? = Check the return is convertable to a truth value

def check_filter(fn, stream):
    try:
        _check_filter(fn, stream)
    except TypeError:
        raise
    except:
        pass

def _check_filter(fn, stream):
    _check_arg_matching_schema(fn, stream)

# Check the callable for a split
#
# - Can be passed a single argument
# - Has an argument type that is compatible the the schema type.
# - TODO? = Check the return is convertable to a int value

def check_split(fn, stream):
    try:
        _check_split(fn, stream)
    except TypeError:
        raise
    except:
        import traceback
        traceback.print_exc()
        pass

def _check_split(fn, stream):
    _check_arg_matching_schema(fn, stream)

# Check the callable for for_each
#
# - Can be passed a single argument
# - Has an argument type that is compatible the the schema type.
# - TODO? = Check the return is nothing?

def check_for_each(fn, stream):
    try:
        _check_for_each(fn, stream)
    except TypeError:
        raise
    except:
        import traceback
        traceback.print_exc()
        pass

def _check_for_each(fn, stream):
    _check_arg_matching_schema(fn, stream)

# Check the callable for map
#
# - Can be passed a single argument
# - Has an argument type that is compatible the the schema type.
# - Infer output schema from type hint

def check_map(fn, stream):
    try:
        return _check_map(fn, stream)
    except TypeError:
        raise
    except:
        import traceback
        traceback.print_exc()
        pass

def _check_map(fn, stream):
    rt_hint = _check_arg_matching_schema(fn, stream)
    if rt_hint:
        return _schema_from_type(rt_hint)

# Check the callable for flat_map
#
# - Can be passed a single argument
# - Has an argument type that is compatible the the schema type.
# - Infer output schema from type hint
# - TODO - Check return type is iterable

def check_flat_map(fn, stream):
    try:
        return _check_flat_map(fn, stream)
    except TypeError:
        raise
    except:
        import traceback
        traceback.print_exc()
        pass

def _get_T(type_):
    if hasattr(type_, '__origin__') and hasattr(type_, '__args__'):
        if len(type_.__args__) == 1:
            return type_.__args__[0]

def _check_flat_map(fn, stream):
    rt_hint = _check_arg_matching_schema(fn, stream)
    if rt_hint:
        et = _get_T(rt_hint)
        if et and typing.Iterable[et] == rt_hint:
            return _schema_from_type(et)

# Check the callable for Window.aggregate
#
# - Can be passed a single argument
# - Has an argument type that is compatible with List[schema type].
# - Infer output schema from type hint

def check_aggregate(fn, window):
    try:
        return _check_aggregate(fn, window)
    except TypeError:
        raise
    except:
        import traceback
        traceback.print_exc()
        pass

def _check_aggregate(fn, stream):
    rt_hint = _check_collection_arg_matching_schema(fn, stream, typing.List, typing.Iterable)
    if rt_hint:
        return _schema_from_type(rt_hint)

# Check the hint for paramter the tuple will be passed as
#  matches the schema.
# If coll_type is set then it is a collection type, e.g. typing.List
# to verify the function can be passed a list of tuples.
#
# Returns the type hint of the function's return if
# it has one AND type_checking is true. Otherwiae return None
def _check_arg_matching_schema(fn, stream):

    if inspect.isroutine(fn):
        n = 1
    elif type(fn) == type:
        # Type is a class hence all returned objects
        # will be of that type
        # TODO argument checking for classes
        #fn = fn.__init__
        #n = 2 # includes self
        return fn
    elif callable(fn):
        fn = fn.__call__
        n = 2 # includes self

    _check_arg_count(fn, 1)

    if stream.topology.type_checking:
        if stream._hints:
            hint, param = _get_arg_hint(fn, n-1)
            if hint:
                source_type = stream._hints.type_
                _check_matching(fn, source_type, hint, param)

        return _fn_return_hint(fn)

def _check_collection_arg_matching_schema(fn, stream, *ctypes):

    if inspect.isroutine(fn):
        n = 1
    elif callable(fn):
        fn = fn.__call__
        n = 2 # includes self

    _check_arg_count(fn, 1)

    if stream.topology.type_checking:
        # TODO aggregate type checking
        if False and stream._hints:
            target, param = _get_arg_hint(fn, n-1)
            if target:
                if target == typing.Any:
                    pass
                else:
                    ok = None
                    source_type = stream._hints.type_
                    for coll_type in ctypes:
                         expected = coll_type[source_type]
                         if target == expected:
                             ok = True ; break
     
                if not ok:
                    if type(target) == type:
                        raise _type_error(fn, target, param, expected)
                    et = _get_T(target)
                    if et:
                        raise _type_error(fn, target, param, expected)

        return _fn_return_hint(fn)

def _type_error(fn, target, param, source):
    return TypeError('Callable {} expects {} for parameter {}, being passed {}' .format(fn.__name__, str(target), param.name, str(source)))

def _check_arg_count(fn, n):

    try:
        sig = inspect.signature(fn)
    except ValueError:
        return
    if len(sig.parameters) == n:
        return

    # Look for default values
    if len(sig.parameters) > n:
        i = 0
        ok = True
        for pn in sig.parameters:
            if i >= n:
                if sig.parameters[pn].default == inspect.Parameter.empty:
                    ok = False
                    break
            i += 1
        if ok:
            return

    raise TypeError('Callable {} takes {} arguments, {} expected' .format(fn, len(sig.parameters), n))


def _schema_from_type(type_):
    try:
        schema = streamsx.topology.schema._normalize(schema=type_, allow_none=False)
    except ValueError:
        schema = None
    return _SchemaHints(type_, schema)

def _get_arg_hint(fn, pos):
    hints = typing.get_type_hints(fn)
    if not hints:
        return None,None

    try:
        sig = inspect.signature(fn)
    except ValueError:
        return None,None
    i = 0
    for pn in sig.parameters:
        i += 1
        if i > pos:
            break
    return hints.get(pn), sig.parameters[pn]

STR_HINTS = _schema_from_type(str)

def _check_matching(fn, source, target, param):

    if target == typing.Any or source == typing.Any:
        return True

    if source == target:
        return True

    try:
        if issubclass(source, target):
            return True
    except TypeError:
        if typing.Optional[source] == target:
            return True
        return None

    raise _type_error(fn, target, param, source)
    #raise TypeError('Callable {} expects {} for parameter {}, being passed {}' .format(fn.__name__, target.__name__, param.name, source.__name__))



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2016,2018
import sys
import sysconfig
import inspect
import imp
import glob
import fcntl
import fnmatch
import os
import shutil
import argparse
import subprocess
import xml.etree.ElementTree as ET
import html
from streamsx.spl.spl import _OperatorType
from streamsx.spl.spl import _valid_op_parameter
import streamsx.spl.spl

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__

# Return the root of the com.ibm.streamsx.topology toolkit
def _topology_tk_dir():
    dir_ = os.path.dirname(__file__) # streamsx/scripts
    dir_ = os.path.dirname(dir_) # streamsx
 
    # See if we are being run from streamsx package
    pkg_tk = os.path.join(dir_, '.toolkit', 'com.ibm.streamsx.topology')
    if os.path.isdir(pkg_tk):
        return pkg_tk

    # Run from the SPL toolkit
    for _ in range(4):
        dir_ = os.path.dirname(dir_)
    return dir_

def replaceTokenInFile(file, token, value):
    f = open(file,'r')
    contents = f.read()
    f.close()

    newcontents = contents.replace(token, value)

    f = open(file,'w')
    f.write(newcontents)
    f.close()

def _optype(opobj):
    if hasattr(opobj, '_splpy_optype'):
        return opobj._splpy_optype
    return None

def _opfile(opobj):
    return opobj._splpy_file

def _opcallable(opobj):
    return opobj._splpy_callable

def _opdoc(opobj):
    return opobj._splpy_docpy

_INFO_XML_TEMPLATE="""<?xml version="1.0" encoding="UTF-8"?>
<toolkitInfoModel
  xmlns="http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo"
  xmlns:common="http://www.ibm.com/xmlns/prod/streams/spl/common">
  <identity>
    <name>__SPLPY_TOOLKIT_NAME__</name>
    <description>Automatic generated toolkit description file.</description>
    <version>1.0.0</version>
    <requiredProductVersion>4.2</requiredProductVersion>
  </identity>
  <dependencies/>
  <resources>
    <messageSet name="TopologySplpyResource">
      <lang default="true">en_US/TopologySplpyResource.xlf</lang>
      <lang>de_DE/TopologySplpyResource.xlf</lang>
      <lang>es_ES/TopologySplpyResource.xlf</lang>
      <lang>fr_FR/TopologySplpyResource.xlf</lang>
      <lang>it_IT/TopologySplpyResource.xlf</lang>
      <lang>ja_JP/TopologySplpyResource.xlf</lang>
      <lang>ko_KR/TopologySplpyResource.xlf</lang>
      <lang>pt_BR/TopologySplpyResource.xlf</lang>
      <lang>ru_RU/TopologySplpyResource.xlf</lang>
      <lang>zh_CN/TopologySplpyResource.xlf</lang>
      <lang>zh_TW/TopologySplpyResource.xlf</lang>
    </messageSet>
  </resources>
</toolkitInfoModel>
"""


# Create SPL operator parameters from the Python class
# (functions cannot have parameters)
# The parameters are taken from the signature of
# the __init__ method. In the spirit of Python
# the default for non-annotated function parameters
# is to map to operator parameters that take any type
# with a cardinality of 1. If the function parameter
# has a default value, then the operator parameter is optional

_OP_PARAM_TEMPLATE ="""
 <parameter>
  <name>__SPLPY__PARAM_NAME__SPLPY__</name>
  <description></description>
  <optional>__SPLPY__PARAM_OPT__SPLPY__</optional>
  <rewriteAllowed>true</rewriteAllowed>
  <expressionMode>AttributeFree</expressionMode>
  <type></type>
  <cardinality>1</cardinality>
 </parameter>"""

_OP_INPUT_PORT_SET_TEMPLATE ="""
<inputPortSet>
  <description>
                                    __SPLPY__INPORT_DESCRIPTION__SPLPY__
  </description>
  <tupleMutationAllowed>false</tupleMutationAllowed>
  <windowingMode>NonWindowed</windowingMode>
  <windowPunctuationInputMode>Oblivious</windowPunctuationInputMode>
  <cardinality>1</cardinality>
  <optional>false</optional>
</inputPortSet>
"""

_OP_OUTPUT_PORT_SET_TEMPLATE ="""
<outputPortSet>
  <description>
    __SPLPY__OUTPORT_DESCRIPTION__SPLPY__
  </description>
  <expressionMode>Nonexistent</expressionMode>
  <autoAssignment>false</autoAssignment>
  <completeAssignment>false</completeAssignment>
  <rewriteAllowed>true</rewriteAllowed>
  <windowPunctuationOutputMode>Free</windowPunctuationOutputMode>
  <tupleMutationAllowed>false</tupleMutationAllowed>
  <cardinality>1</cardinality>
  <optional>false</optional>
</outputPortSet>
"""





class _Extractor(object):
    def __init__(self, args):
        self._cmd_args = self._parse_cmd_args(args)
        self._tk_dir = self._cmd_args.directory

    def _parse_cmd_args(self, args):
        cmd_parser = argparse.ArgumentParser(description='Extract SPL operators from decorated Python classes and functions.')
        cmd_parser.add_argument('-i', '--directory', required=True,
                   help='Toolkit directory')
        cmd_parser.add_argument('--make-toolkit', action='store_true',
                   help='Index toolkit using spl-make-toolkit')
        cmd_parser.add_argument('-v', '--verbose', action='store_true',
                   help='Print more diagnostics')
        return cmd_parser.parse_args(args)

    def _make_namespace_dir(self, ns):
         nsdir = os.path.join(self._tk_dir, ns)
         if os.path.isdir(nsdir):
             return nsdir
         os.mkdir(nsdir)
         return nsdir

    def _make_operator_dir(self, nsdir, name):
         oppath = os.path.join(nsdir, name)
         if (os.path.isdir(oppath)):
             shutil.rmtree(oppath)
         os.mkdir(oppath)
         return oppath

    def _make_toolkit(self):
        # Now make the toolkit if required
        if self._cmd_args.make_toolkit:
            si = os.environ['STREAMS_INSTALL']
            mktk = os.path.join(si, 'bin', 'spl-make-toolkit')
            mktk_args = [mktk, '--directory', self._cmd_args.directory, '--make-operator']
            subprocess.check_call(mktk_args)

    # Process python objects in a module looking for SPL operators
    # dynm - introspection for the modeul
    # module - module name
    # ops - list of potential operators (functions)
    def _process_operators(self, dynm, module, streams_python_file, ops):
        for opname, opobj in ops:
            if inspect.isbuiltin(opobj):
                continue
            if opname.startswith('spl'):
                continue
            optype = _optype(opobj)
            if optype is None:
                continue
            if optype == _OperatorType.Ignore:
                continue
            if streams_python_file != _opfile(opobj):
                continue
            self._common_tuple_operator(dynm, module, opname, opobj)

    def _copy_globalization_resources(self):
        '''Copy the language resource files for python api functions
    
        This function copies the TopologySplpy Resource files from Topology toolkit directory
        into the impl/nl folder of the project.
        Returns: the list with the copied locale strings'''
        rootDir = os.path.join(_topology_tk_dir(), "impl", "nl")
        languageList = []
        for dirName in os.listdir(rootDir):
            srcDir = os.path.join(_topology_tk_dir(), "impl", "nl", dirName)
            if (os.path.isdir(srcDir)) and (dirName != "include"):
                dstDir = os.path.join(self._tk_dir, "impl", "nl", dirName)
                try:
                    print("Copy globalization resources " + dirName)
                    os.makedirs(dstDir)
                except OSError as e:
                    if (e.errno == 17) and (os.path.isdir(dstDir)):
                        if self._cmd_args.verbose:
                            print("Directory", dstDir, "exists")
                    else:
                        raise
                srcFile = os.path.join(srcDir, "TopologySplpyResource.xlf")
                if os.path.isfile(srcFile):
                    res = shutil.copy2(srcFile, dstDir)
                    languageList.append(dirName)
                    if self._cmd_args.verbose:
                        print("Written: " + res)
        return languageList

    #
    # module - module for operator
    # opname - name of the SPL operator
    # opobj - decorated object defining operator
    #
    def _common_tuple_operator(self, dynm, module, opname, opobj) :        
        if (not hasattr(dynm, 'spl_namespace')) and hasattr(dynm, 'splNamespace'):
            ns = getattr(dynm, 'splNamespace')()
        else:
            ns = getattr(dynm, 'spl_namespace')()
        print(ns + "::" + opname)
        # Print the summary of the class/function
        _doc = inspect.getdoc(opobj)
        if _doc is not None:
            _doc = _doc.splitlines()[0]
            print("  ", _doc)
        nsdir = self._make_namespace_dir(ns)
        opdir = self._make_operator_dir(nsdir, opname)
        self._copy_template_dir("common")
        self._copy_template_dir("icons")
        self._copy_python_dir("packages")
        self._copy_python_dir("include")
        self._copy_CGT(opdir, ns, opname, opobj)
        self._write_config(dynm, opdir, module, opname, opobj)

    def _create_op_parameters(self, opmodel_xml, name, opObj):
        opparam_xml = ''
        if _opcallable(opObj) == 'class':
            pmds = init_sig = inspect.signature(opObj._splpy_wrapped.__init__).parameters
            itpmds = iter(pmds)
            # first argument to __init__ is self (instance ref)
            next(itpmds)
        
            for pn in itpmds:
                pmd = pmds[pn]
                _valid_op_parameter(pn)
                px = _OP_PARAM_TEMPLATE
                px = px.replace('__SPLPY__PARAM_NAME__SPLPY__', pn)
                px = px.replace('__SPLPY__PARAM_OPT__SPLPY__', 'false' if pmd.default== inspect.Parameter.empty else 'true' )
                opparam_xml = opparam_xml + px
        replaceTokenInFile(opmodel_xml, '__SPLPY__PARAMETERS__SPLPY__', opparam_xml)
    
    def _copy_CGT(self, opdir, ns, name, funcTuple):
         cgtbase = _optype(funcTuple).spl_template
         optemplate = os.path.join(_topology_tk_dir(), "opt", "python", "templates","operators", cgtbase)
         opcgt_cpp = os.path.join(opdir, name + '_cpp.cgt')
         shutil.copy(optemplate + '_cpp.cgt', opcgt_cpp)
         shutil.copy(optemplate + '_h.cgt', os.path.join(opdir, name + '_h.cgt'))
         opmodel_xml = os.path.join(opdir, name + '.xml')
         shutil.copy(optemplate + '.xml', opmodel_xml)
         replaceTokenInFile(opmodel_xml, "__SPLPY__MAJOR_VERSION__SPLPY__", str(sys.version_info.major));
         replaceTokenInFile(opmodel_xml, "__SPLPY__MINOR_VERSION__SPLPY__", str(sys.version_info.minor));
         self._create_op_parameters(opmodel_xml, name, funcTuple)
         self._create_op_spldoc(opmodel_xml, name, funcTuple)
         if cgtbase == 'PythonPrimitive':
             self._create_primitive(opmodel_xml, name, funcTuple)
         else:
             self._create_ip_spldoc(opmodel_xml, name, funcTuple)

    ## Create SPL doc entries in the Operator model xml file.
    ##
    def _create_op_spldoc(self, opmodel_xml, name, opobj):
         opdoc = inspect.getdoc(opobj)
         if opdoc is None:
             opdoc = 'Callable: ' + name + "\n"

         opdoc = html.escape(opdoc)

         # Optionally include the Python source code
         if _opdoc(opobj):
             decor = None
             if inspect.isclass(opobj):
                 if hasattr(opobj, '_splpy_decor'):
                      decor = opobj._splpy_decor
                 opobj = inspect.getmro(opobj)[1]
             try:
                 _pysrc = inspect.getsource(opobj)
                 opdoc += "\n"
                 opdoc += "# Python\n";

                 if decor:
                     opdoc += '    '
                     opdoc += decor
                     opdoc += '\n'

                 for _line in str.splitlines(_pysrc):
                     opdoc += '    '
                     opdoc += html.escape(_line)
                     opdoc += '\n'
             except:
                 pass
         
         replaceTokenInFile(opmodel_xml, "__SPLPY__DESCRIPTION__SPLPY__", opdoc);

    def _create_ip_spldoc(self, opmodel_xml, name, opobj):
         if opobj._splpy_style == 'dictionary':
           _p0doc = """
           Tuple attribute values are passed by name to the Python callable using `\*\*kwargs`.
                 """
         elif opobj._splpy_style == 'tuple':
           _p0doc = """
           Tuple attribute values are passed by position to the Python callable.
                 """
         else:
           _p0doc = ''
     
         replaceTokenInFile(opmodel_xml, "__SPLPY__INPORT_0_DESCRIPTION__SPLPY__", _p0doc);

    def _create_primitive(self, opmodel_xml, name, cls):
        # input ports
        if cls._splpy_input_ports:
            ipd = ''
            for portfn in cls._splpy_input_ports:
                tip = _OP_INPUT_PORT_SET_TEMPLATE
                fndoc = inspect.getdoc(portfn)
                if not fndoc:
                    fndoc = portfn.__name__
                tip = tip.replace('__SPLPY__INPORT_DESCRIPTION__SPLPY__', fndoc)
                ipd += tip
        else:
            ipd = "<!-- no input ports -->"
        replaceTokenInFile(opmodel_xml, "__SPLPY__INPORTS__SPLPY__", ipd);

        # output ports
        if cls._splpy_output_ports:
            opd = ''
            for portfn in cls._splpy_output_ports:
                top = _OP_OUTPUT_PORT_SET_TEMPLATE
                opd += top
        else:
            opd = "<!-- no output ports -->"
        replaceTokenInFile(opmodel_xml, "__SPLPY__OUTPORTS__SPLPY__", opd);
   
    # Write information about the Python function parameters.
    #
    def _write_style_info(self, cfgfile, opobj):
            style = opobj._splpy_style

            if isinstance(style, list):
                pm_style = '(' + ','.join(["'{0}'".format(_) for _ in opobj._splpy_style]) + ')'
            else:
                pm_style = "'" + style + "'"
            cfgfile.write('sub splpy_ParamStyle {'+ pm_style + '}\n')
             
            if isinstance(opobj._splpy_fixed_count, list):
                pm_fixed = '(' + ','.join(["'{0}'".format(_) for _ in opobj._splpy_fixed_count]) + ')'
            else:
                pm_fixed = "'" + str(opobj._splpy_fixed_count) + "'"

            cfgfile.write('sub splpy_FixedParam { '+ pm_fixed + "}\n")
 

    # Write out the configuration for the operator
    # as a set of Perl functions that return useful values
    # for the code generator
    def _write_config(self, dynm, opdir, module, opname, opobj):
        cfgpath = os.path.join(opdir, 'splpy_operator.pm')
        cfgfile = open(cfgpath, 'w')
        cfgfile.write('sub splpy_Module { \''+ module   + "\'}\n")
        cfgfile.write('sub splpy_OperatorCallable {\'' + _opcallable(opobj) + "\'}\n")
        cfgfile.write('sub splpy_FunctionName {\'' + opname + "\'}\n")
        cfgfile.write('sub splpy_OperatorType {\'' + _optype(opobj).name + "\'}\n")

        cfgfile.write('sub splpy_PyMajor {' + str(sys.version_info.major) + "}\n")
        cfgfile.write('sub splpy_PyMinor {' + str(sys.version_info.minor) + "}\n")
        self._write_style_info(cfgfile, opobj)

        if hasattr(dynm, 'spl_pip_packages'):
            pp = getattr(dynm, 'spl_pip_packages')()
            if not isinstance(pp, list):
                pp = list(pp)
        else:
            pp = []

        cfgfile.write('sub splpy_Packages {(' + ','.join(["'{0}'".format(_) for _ in pp]) + ')}\n')

        cfgfile.write("1;\n")
        cfgfile.close()

    # Copy a single file from the templates directory to the newly created operator directory
    def _copy_template_dir(self, dir_):
        self._copy_python_dir(os.path.join("templates", dir_))

    def _copy_python_dir(self, dir_):
        cmn_src = os.path.join(_topology_tk_dir(), "opt", "python", dir_);
        cmn_dst = os.path.join(self._tk_dir, "opt", ".splpy", os.path.basename(dir_))
        if os.path.isdir(cmn_dst):
            shutil.rmtree(cmn_dst)
        # Remove the old directory used - see #1997
        old_dst = os.path.join(self._tk_dir, "opt", ".__splpy")
        if os.path.isdir(old_dst):
            shutil.rmtree(old_dst)
        shutil.copytree(cmn_src, cmn_dst)

    def _setup_info_xml(self, languageList):
        '''Setup the info.xml file
        
        This function prepares or checks the info.xml file in the project directory
        - if the info.xml does not exist in the project directory, it copies the template info.xml into the project directory.
          The project name is obtained from the project directory name
        - If there is a info.xml file, the resource section is inspected. If the resource section has no valid message set
          description for the TopologySplpy Resource a warning message is printed'''
        infoXmlFile = os.path.join(self._tk_dir, 'info.xml')
        print('Check info.xml:', infoXmlFile)
        try:
            TopologySplpyResourceMessageSetFound = False
            TopologySplpyResourceLanguages = []
            tree = ET.parse(infoXmlFile)
            root = tree.getroot()
            for resources in root.findall('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}resources'):
                if self._cmd_args.verbose:
                    print('Resource: ', resources.tag)
                for messageSet in resources.findall('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}messageSet'):
                    if self._cmd_args.verbose:
                        print('Message set:', messageSet.tag, messageSet.attrib)
                    if 'name' in messageSet.attrib:
                        if messageSet.attrib['name'] == 'TopologySplpyResource':
                            TopologySplpyResourceMessageSetFound = True
                            for lang in messageSet.findall('{http://www.ibm.com/xmlns/prod/streams/spl/toolkitInfo}lang'):
                                language = os.path.dirname(lang.text)
                                TopologySplpyResourceLanguages.append(language)
            if TopologySplpyResourceMessageSetFound:
                TopologySplpyResourceLanguages.sort()
                languageList.sort()
                copiedLanguagesSet = set(languageList)
                resourceLanguageSet = set(TopologySplpyResourceLanguages)
                if self._cmd_args.verbose:
                    print('copied language resources:\n', languageList)
                    print('TopologySplpyResource from info.xml:\n', TopologySplpyResourceLanguages)
                if copiedLanguagesSet == resourceLanguageSet:
                    print('Resource section of info.xml verified')
                else:
                    errstr = """"ERROR: Message set for the "TopologySplpyResource" is incomplete or invalid. Correct the resource section in info.xml file.
    
                    Sample info xml:\n""" + _INFO_XML_TEMPLATE
                    sys.exit(errstr)
            else:
                errstr = """"ERROR: Message set for the "TopologySplpyResource" is missing. Correct the resource section in info.xml file.
    
                    Sample info xml:\n""" + _INFO_XML_TEMPLATE
                sys.exit(errstr)
        except FileNotFoundError as e:
            print("WARNING: File info.xml not found. Creating info.xml from template")
            #Get default project name from project directory
            projectRootDir = os.path.abspath(self._tk_dir) #os.path.abspath returns the path without trailing /
            projectName = os.path.basename(projectRootDir)
            infoXml=_INFO_XML_TEMPLATE.replace('__SPLPY_TOOLKIT_NAME__', projectName)
            f = open(infoXmlFile, 'w')
            f.write(infoXml)
            f.close()
        except SystemExit as e:
            raise e
        except:
            errstr = """ERROR: File info.xml is invalid or not accessible
    
                Sample info xml:\n""" + _INFO_XML_TEMPLATE
            sys.exit(errstr)

def _extract_from_toolkit(args):
    """
    Look at all the modules in opt/python/streams (opt/python/streams/*.py)
    and extract any spl decorated function as an operator.
    """

    extractor = _Extractor(args)
    if extractor._cmd_args.verbose:
        print("spl-python-extract:", __version__)
        print("Topology toolkit location:", _topology_tk_dir())

    tk_dir = extractor._tk_dir

    tk_streams = os.path.join(tk_dir, 'opt', 'python', 'streams')
    if not os.path.isdir(tk_streams) or not fnmatch.filter(os.listdir(tk_streams), '*.py'):
        # Nothing to do for Python extraction
        extractor._make_toolkit()
        return

    lf = os.path.join(tk_streams, '.lockfile')
    with  open(lf, 'w') as lfno:
        fcntl.flock(lfno, fcntl.LOCK_EX)

        tk_idx = os.path.join(tk_dir, 'toolkit.xml')
        tk_time = os.path.getmtime(tk_idx) if os.path.exists(tk_idx) else None
        changed = False if tk_time else True
        if tk_time:
            for mf in glob.glob(os.path.join(tk_streams, '*.py')):
               if os.path.getmtime(mf) >= tk_time:
                   changed = True
                   break

        if changed:
            path_items = _setup_path(tk_dir, tk_streams)

            for mf in glob.glob(os.path.join(tk_streams, '*.py')):
                print('Checking ', mf, 'for operators')
                name  = inspect.getmodulename(mf)
                dynm = imp.load_source(name, mf)
                streams_python_file = inspect.getsourcefile(dynm)
                extractor._process_operators(dynm, name, streams_python_file, inspect.getmembers(dynm, inspect.isfunction))
                extractor._process_operators(dynm, name, streams_python_file, inspect.getmembers(dynm, inspect.isclass))

            langList = extractor._copy_globalization_resources()
            if extractor._cmd_args.verbose:
                print("Available languages for TopologySplpy resource:", langList)
            extractor._setup_info_xml(langList)

            extractor._make_toolkit()

            _reset_path(path_items)
            fcntl.flock(lfno, fcntl.LOCK_UN)

def _setup_path(tk_dir, tk_streams):
    sys.path.insert(1, tk_streams)
    items = [tk_streams]
    tk_packages = os.path.join(tk_dir, 'opt', 'python', 'packages')
    if os.path.isdir(tk_packages):
        sys.path.insert(1, tk_packages)
        items.append(tk_packages)
    tk_modules = os.path.join(tk_dir, 'opt', 'python', 'modules')
    if os.path.isdir(tk_modules):
        sys.path.insert(1, tk_modules)
        items.append(tk_modules)
    return items

def _reset_path(items):
    for p in items:
        sys.path.remove(p)

def main(args=None):
    streamsx._streams._version._mismatch_check('streamsx.spl.spl')
    try:
        streamsx.spl.spl._EXTRACTING = True
        _extract_from_toolkit(args)
    finally:
        streamsx.spl.spl._EXTRACTING = False

if __name__ == '__main__':
    main()



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2019,2020

"""
Composite transformations.

.. versionadded:: 1.14
"""


__all__ = ['Composite', 'Source', 'Map', 'ForEach']

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__

from abc import ABC, abstractmethod
from typing import Optional
import streamsx.typing
import streamsx.topology.topology

class Composite(ABC):
    """
    Composite transformations support a single logical transformation
    being a composite of one or more basic transformations.

    A composite transformation is implemented as a sub-class
    of :py:class:`Source`, :py:class:`Map` or :py:class:`ForEach`
    whose ``populate`` method populates the topology with the
    required basic transformations. For example a ``Source`` composite
    might have use :py:meth:`~Topology.source` followed by a
    :py:meth:`~Stream.filter` to filter out unwanted events
    and then a :py:meth:`~Stream.map` to parse the event into
    a structured schema.

    Composites may use other composites during ``populate``.


    Composites can control how the basic transformations are
    visually represented. By default any transformations within
    a composite are grouped visually. A composite may alter this using these
    attributes of the composite instance:

        * ``kind`` - Sets the name of operator kind for a group or single operator.  Defaults to a combination of the module and class name of the composite, e.g. ``streamsx.standard.utility::Sequence``. Set to a false value to disable any modification of the visual representation of the composite's transformations.
        * ``group`` - Set to a false value to disable any grouping of multiple transformations. Defaults to ``True`` to enable grouping.

    The values of ``kind`` and ``group`` are checked after the expansion
    of the composite using ``populate``.

    """
    @classmethod
    def _check_type(cls, obj, type_):
        if not isinstance(obj, type_):
            raise TypeError()

    def _group_ops(self, topology, name):
        if not hasattr(self, 'group'):
            self.group = True
        if not hasattr(self, 'kind'):
            ns = self.__class__.__module__
            if ns == '__main__':
                self.kind = self.__class__.__name__
            else:
                self.kind = ns + '::' + self.__class__.__name__

        if not name:
            if '::' in self.kind:
                name = self.kind.split('::')[-1]
            else:
                name = self.kind
        return _GroupOps(self, topology, name)


class _GroupOps(object):
    def __init__(self, composite, topology, name):
        self.composite = composite
        self.topology = topology
        self.name = name
    def __enter__(self):
        self.ops = len(self.topology.graph.operators)
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
           new_ops = len(self.topology.graph.operators) - self.ops
           if new_ops > 1 and self.composite.group and self.composite.kind:
               self._grouping()
           if new_ops == 1 and self.composite.kind:
               self._single()

    def _grouping(self):
        group_id = None
        for i in range(self.ops, len(self.topology.graph.operators)):
            op = self.topology.graph.operators[i]
            group_id = op._layout_group(self.composite.kind, self.name, group_id=group_id)

    def _single(self):
        op = self.topology.graph.operators[self.ops]
        op._layout(kind=self.composite.kind, name=op.name, orig_name=self.name)
 

class Source(Composite):
    """
    Abstract composite source.

    An instance of a subclass can be passed to :py:meth:`~Topology.source`
    to create a source stream that is composed of one or more basic transformations.

    Example assuming ``RawTweets`` is Python iterable that produces
    raw tweets::

        class Tweets(streamsx.topology.composite.Source):
            def __init__(self, track):
                self.track = track
     
            def populate(self, topology, name, **options):
                # get all the tweets
                tweets = topology.source(RawTweets(track=self.track), name=name)
                # filter so that only with a message are returned
                return tweets.filter(lambda tweet : tweet['text'])

    This class can then be used as follows::

        topo = Topology()
        gf_tweets = topo.source(Tweets(track=['glutenfree', 'gf']))

    """

    def _add(self, topology, name, **options):
        with self._group_ops(topology, name):
            s = self.populate(topology, name, **options)
        Source._check_type(s, streamsx.topology.topology.Stream)
        return s

    @abstractmethod
    def populate(self, topology:streamsx.topology.topology.Topology, name:Optional[str], **options) -> streamsx.topology.topology.Stream:
        """
        Populate the topology with this composite source.

        Args:
            topology: Topology containing the source.
            name: Name passed into ``source``.
            **options: Future options passed to ``source``.

        Returns:
            Stream: Single stream representing the source.
        """
        pass


class Map(Composite):
    """
    Abstract composite map transformation.

    An instance of a subclass can be passed to :py:meth:`~streamsx.topology.topology.Stream.map`
    to create a stream that is composed of one or more basic transformations
    of an input stream.

    Example::

        class WordCount(streamsx.topology.composite.Map):
            def __init__(self, period, update):
                self.period = period
                self.update = update
     
            def populate(self, topology, stream, schema, name, **options):
                words = stream.flat_map(lambda line : line.split())
                win = words.last(size=self.period).trigger(self.update).partition(lambda s : s)
                return win.aggregate(lambda values : (values[0], len(values)))
    """

    def _add(self, stream, schema, name, **options):
        with self._group_ops(stream.topology, name):
            s = self.populate(stream.topology, stream, schema, name, **options)
        Map._check_type(s, streamsx.topology.topology.Stream)
        if schema:
            s = s.map(schema=schema)
        return s

    @abstractmethod
    def populate(self, topology:streamsx.topology.topology.Topology, stream:streamsx.topology.topology.Stream, schema:streamsx.typing.Schema, name:Optional[str], **options) -> streamsx.topology.topology.Stream:
        """
        Populate the topology with this composite map transformation.

        Args:
            topology: Topology containing the composite map.
            stream: Stream to be transformed.
            schema: Schema passed into ``map``.
            name: Name passed into ``map``.
            **options: Future options passed to ``map``.

        Returns:
            Stream: Single stream representing the transformation of `stream`.
        """
        pass

class ForEach(Composite):
    """
    Abstract composite for each transformation.

    An instance of a subclass can be passed to :py:meth:`~streamsx.topology.topology.Stream.for_each` to create a sink (stream termination) that is
    composed of one or more basic transformations of an input stream.
    """

    def _add(self, stream, name, **options):
        with self._group_ops(stream.topology, name):
            s = self.populate(stream.topology, stream, name, **options)
        ForEach._check_type(s, streamsx.topology.topology.Sink)
        return s

    @abstractmethod
    def populate(self, topology:streamsx.topology.topology.Topology, stream:streamsx.topology.topology.Stream, name:Optional[str], **options) -> streamsx.topology.topology.Sink:
        """
        Populate the topology with this composite for each transformation.

        Args:
            topology: Topology containing the composite map.
            stream: Stream to be transformed.
            name: Name passed into ``for_each``.
            **options: Future options passed to ``for_each``.

        Returns:
            ~streamsx.topology.topology.Sink: Termination for this composite transformation of `stream`.
        """
        pass



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2016,2019
"""
Schemas for streams.

********
Overview
********

A stream represents an unbounded flow of tuples with a declared schema so that each tuple on the stream complies with the schema. A stream's schema may be one of:

    * :py:class:`~StreamsSchema` structured schema - a tuple is a sequence of attributes, and an attribute is a named value of a specific type.
    * :py:const:`~CommonSchema.Json` a tuple is a JSON object.
    * :py:const:`~CommonSchema.String` a tuple is a string.
    * :py:const:`~CommonSchema.Python` a tuple is any Python object, effectively an untyped stream.

******************
Structured schemas
******************

A structured schema is a sequence of attributes, and an attribute is a named value of a specific type. For example a stream of sensor readings can be represented as a schema with three attributes ``sensor_id``, ``ts`` and ``reading`` with types of ``int64``, ``int64`` and ``float64`` respectively.

This schema can be declared a number of ways:

Python 3.6::

    class SensorReading(typing.NamedTuple):
        sensor_id: int
        ts: int
        reading: float

    sensors = raw_readings.map(parse_sensor, schema=SensorReading)
    
Python 3::

    SensorReading = typing.NamedTuple('SensorReading',
        [('sensor_id', int), ('ts', int), ('reading', float)]

    sensors = raw_readings.map(parse_sensor, schema=SensorReading)

Python 3::

    sensors = raw_readings.map(parse_sensor,
        schema='tuple<int64 sensor_id, int64 ts, float64 reading>')

The supported types are defined by IBM Streams and are listed in :py:class:`~StreamSchema`.

Structured schemas provide type-safety and efficient network serialization when compared to passing a ``dict`` using :py:const:`~CommonSchema.Python` streams. 

Streams with structured schemas can be interchanged with any IBM Streams application using :py:meth:`~streamsx.topology.topology.Stream.publish` and :py:meth:`~streamsx.topology.topology.Topology.subscribe` maintaining type safety.

**************************
Defining a stream's schema
**************************

Every stream within a :py:class:`~streamsx.topology.topology.Topology` has defined schema. The schema may be defined explictly (for example :py:meth:`~streamsx.topology.topology.Stream.map` or :py:meth:`~streamsx.topology.topology.Topology.subscribe`) or implicity (for example :py:meth:`~streamsx.topology.topology.Stream.filter` produces a stream with the same schema as its input stream).

Explictly defining a stream's schema is flexible and various types of values are accepted as the schema.

    * Builtin types as aliases for common schema types:

        * ``json`` (module) - for  :py:const:`~CommonSchema.Json`
        * ``str`` - for  :py:const:`~CommonSchema.String`
        * ``object`` - for :py:const:`~CommonSchema.Python`

    * Values of the enumeration :py:class:`CommonSchema`
    * An instance of ``typing.NamedTuple`` (Python 3)
    * An instance of :py:class:`StreamSchema`
    * A string of the format ``tuple<...>`` defining the attribute names and types. See :py:class:`StreamSchema` for details on the format and types supported.
    * A string containing a namespace qualified SPL stream type (e.g. ``com.ibm.streams.geospatial::FlightPathEncounterTypes.Observation3D``)

"""

# For style dicts passed into Python from Streams C++
# are raw dicts since they are created by Python C-API code
# not the future dict in Python 2.7.
_spl_dict = dict
_spl_object = object

__all__ = ['is_common', 'StreamSchema', 'CommonSchema']

import collections
import decimal
import datetime
import enum
import io
import itertools
import json
import sys
import token
import tokenize
import typing

import streamsx._streams._version
__version__ = streamsx._streams._version.__version__


_spl_str = str

def _normalize(schema, allow_none=True):
    """
    Normalize a schema.
    """
    if allow_none and schema is None:
        return schema
    if isinstance(schema, CommonSchema):
        return schema
    if isinstance(schema, StreamSchema):
        return schema

    if isinstance(schema, str):
        return StreamSchema(schema)

    py_types = {
        _spl_object: CommonSchema.Python,
        _spl_str: CommonSchema.String,
        json: CommonSchema.Json,
        }

    if schema in py_types:
        return py_types[schema]

    # Allow a named tuple with type hints
    # to be used as a schema definition
    if _is_typedtuple(schema):
        return _from_named_tuple(schema)

    raise ValueError("Unknown stream schema type:" + str(schema))

def is_common(schema):
    """
    Is `schema` an common schema.

    Args:
        schema: Scheme to test.

    Returns:
        bool: ``True`` if schema is a common schema, otherwise ``False``.

    """
    if isinstance(schema, StreamSchema):
        return schema.schema() in _SCHEMA_COMMON
    if isinstance(schema, CommonSchema):
        return True
    if isinstance(schema, str):
        return is_common(StreamSchema(schema))
    return False

def _is_namedtuple(cls):
    return cls != tuple and isinstance(cls, type) and issubclass(cls, tuple) and hasattr(cls, '_fields')

def _is_typedtuple(cls):
    return _is_namedtuple(cls) and hasattr(cls, '_field_types')

_SCHEMA_PENDING = '<pending>'
def _is_pending(schema):
    return isinstance(schema, StreamSchema) and schema.schema() == _SCHEMA_PENDING

# Parses a schema of the form 'tuple<...>'
# _parse returns a list of the schema attributes,
# each attribute is a python tuple of:
# (type, name)
# with type being
#    primitive type (str), e.g. 'int32'
#    collection type (tuple), e.g. ('list', 'int32')
#    optional type (tuple), e.g. ('optional', 'int32')
#    nested tuple type (tuple), e.g. ('tuple', [('int32', 'a'), ('float64', 'b')])
# This is an internal api.
#
class _SchemaParser(object):
    """Class for parsing schemas."""
    _SPL_PRIMITIVE_TYPES = { 'boolean',
                         'uint8', 'uint16', 'uint32', 'uint64',
                         'int8', 'int16', 'int32', 'int64',
                         'float32', 'float64',
                         'complex32', 'complex64',
                         'decimal32', 'decimal64', 'decimal128',
                         'rstring', 'ustring',
                         'timestamp', 'blob', 'xml'}

    _SPL_COLLECTION_TYPES = { 'list', 'set'}

    def __init__(self, schema):
        self.schema = schema
        self._type = []

    def _parse_error(self, token):
        raise SyntaxError("Invalid schema:" + self.schema + " token " + str(token))

    def _req_op(self, which):
        token = next(self.tokens)
        if token[0] != tokenize.OP or which != token[1]:
            self._parse_error(token)

    def _parse(self):
        schema = self.schema.replace(">>", ' > > ')
        schema = schema.replace('<<', ' < < ')

        ios = io.StringIO(schema).readline
        self.tokens = tokenize.generate_tokens(ios)
        self._parse_tuple(self._type, next(self.tokens))
        endtoken = next(self.tokens)
        # Python 3.6.7 adds a newline at the end! Issue #1959
        if endtoken[0] == token.NEWLINE:
            endtoken = next(self.tokens)
        if not endtoken[0] == token.ENDMARKER:
            self._parse_error(endtoken)
        return self._type

    def _parse_tuple(self, _type, token):
        if token[0] != tokenize.NAME or 'tuple' != token[1]:
            self._parse_error(token)
        self._req_op('<')
    
        token = None
        while True:
            token = next(self.tokens)
            if token[0] == tokenize.OP:
                if token[1] == ',':
                    continue
                if token[1] == '>':
                    break
                self._parse_error(token)

            if token[0] == tokenize.NAME:
                self._parse_attribute_type(_type, token)
                continue

            self._parse_error(token)

    def _parse_type(self, attr_type):
        if attr_type[0] != tokenize.NAME:
            self._parse_error(attr_type)

        if 'tuple' == attr_type[1]:
            nested_tuple = []
            self._parse_tuple(nested_tuple, attr_type)
            return ('tuple', nested_tuple)

        if 'map' == attr_type[1]:
            self._req_op('<')
            key_type = self._parse_type(next(self.tokens))
            self._req_op(',')
            value_type = self._parse_type(next(self.tokens))
            self._req_op('>')
            bound = self._parse_optional_bounded()
            return ('map', (key_type, value_type), bound)

        if 'optional' == attr_type[1]:
            self._req_op('<')
            value_type = self._parse_type(next(self.tokens))
            self._req_op('>')
            return ('optional', value_type)
            
        if attr_type[1] in _SchemaParser._SPL_PRIMITIVE_TYPES:
            if attr_type[1] == 'rstring':
                bound = self._parse_optional_bounded()
                if bound is not None:
                    return 'rstring' + bound
            return attr_type[1]

        if attr_type[1] in _SchemaParser._SPL_COLLECTION_TYPES:
            self._req_op('<')
            element_type = self._parse_type(next(self.tokens))
            self._req_op('>')
            bound = self._parse_optional_bounded()
            return (attr_type[1], element_type, bound)

        self._parse_error(attr_type)

    def _parse_attribute_type(self, _type, attr_type):
        if attr_type[0] != tokenize.NAME:
            self._parse_error(attr_type)

        attr_type = self._parse_type(attr_type)

        attr = (attr_type, self._parse_attribute_name())
        _type.append(attr)

    def _parse_attribute_name(self):
        attr_name = next(self.tokens)
        if attr_name[0] != tokenize.NAME:
            self._parse_error(attr_name)
        return attr_name[1]

    def _parse_optional_bounded(self):
        token = next(self.tokens)
        if token[0] == tokenize.OP and '[' == token[1]:
            bound_info = next(self.tokens)
            if bound_info[0] != tokenize.NUMBER:
                self._parse_error(bound_info)
            bound = str(int(bound_info[0]))
            self._req_op(']')
            return bound
        else:
            # push back the token
            self.tokens = itertools.chain([token], self.tokens)
            return None

def _attribute_names(types):
    names = []
    for attr in types:
        names.append(attr[1])
    return names

# Returns a schema's types as a list of (name, python type) tuples
# suitable for use in creating a typing.NamedTuple
def _attribute_pytypes(types):
    pytypes = []
    for attr in types:
        pytypes.append((attr[1], _type_from_spl(attr[0])))
    return pytypes

_SCHEMA_PYTHON_OBJECT = 'tuple<blob __spl_po>'
_SCHEMA_STRING = 'tuple<rstring string>'
_SCHEMA_JSON = 'tuple<rstring jsonString>'
_SCHEMA_BINARY = 'tuple<blob binary>' # not yet supported
_SCHEMA_XML = 'tuple<xml document>' # not yet supported

_SCHEMA_COMMON = frozenset([_SCHEMA_PYTHON_OBJECT, _SCHEMA_JSON, _SCHEMA_STRING, _SCHEMA_BINARY, _SCHEMA_XML])

_SCHEMA_COMMON_STYLES = {_SCHEMA_PYTHON_OBJECT:_spl_object, _SCHEMA_STRING: _spl_str, _SCHEMA_JSON: _spl_dict, _SCHEMA_BINARY:None, _SCHEMA_XML: None }

class StreamSchema(object) :
    """Defines a schema for a structured stream.

    On a structured stream a tuple is a sequence of attributes,
    and an attribute is a named value of a specific type.

    The supported types are defined by IBM Streams and include such
    types as `int8`, `int16`, `rstring` and `list<float32>`.

    A structured schema can be defined using a ``typing.NamedTuple`` in
    Python 3, a string with the syntax ``tuple<type name [,...]>`` or
    an instance of this class.

    typing.NamedTuple:
 
        A ``typing.NamedTuple`` can be used to define a structured
        schema with the field names and types mapping to the
        structured schema attribute names and types.

        Python types are mapped to IBM Streams types as follows:

        ================================== ================
        Python type                        IBM Streams type
        ================================== ================
        ``str``                            ``rstring``
        ``bool``                           ``boolean``
        ``int``                            ``int64``
        ``float``                          ``float64``
        ``decimal.Decimal``                ``decimal128``
        ``complex``                        ``complex64``
        ``bytes``                          ``blob``
        ``streamsx.spl.types.Timestamp``   ``timestamp``
        ``datetime.datetime``              ``timestamp``
        ``typing.List[T]``                 ``list<T>``
        ``typing.Set[T]``                  ``set<T>``
        ``typing.Mapping[K,V]``            ``map<K,V>``
        ``typing.Optional[T]``             ``optional<T>``
        ================================== ================

        .. note::
            Tuples on a stream with a schema defined by a
            ``typing.NamedTuple`` instance are passed into callables
            as instance of a named tuple with the the correct field
            names and types. This is not guaranteed to be the same class
            instance as the one used to declare the schema.
            

    Tuple string:

        A string of the format `tuple<type name [,...]>` can be used
        to define a structured schema, where `type` is an IBM Streams type.

        Example::

            tuple<rstring id, timestamp ts, float64 value>

        represents a schema with three attributes suitable for a sensor reading.

    IBM Streams types:

    ============================  ==============================  =========================================  =======================================
    Type                          Description                     Python representation                      Conversion from Python
    ============================  ==============================  =========================================  =======================================
    ``boolean``                   True or False                   ``bool``                                   ``bool(value)``
    ``int8``                      8-bit signed integer            ``int``                                    ``int(value)`` truncated to 8 bits
    ``int16``                     16-bit signed integer           ``int``                                    ``int(value)`` truncated to 16 bits
    ``int32``                     32-bit signed integer           ``int``                                    ``int(value)`` truncated to 32 bits
    ``int64``                     64-bit signed integer           ``int``                                    ``int(value)``
    ``uint8``                     8-bit unsigned integer          ``int``                                    -
    ``uint16``                    16-bit unsigned integer         ``int``                                    -
    ``uint32``                    32-bit unsigned integer         ``int``                                    -
    ``uint64``                    64-bit unsigned integer         ``int``                                    -
    ``float32``                   32-bit binary floating point    ``float``                                  ``float(value)`` truncated to 32 bits
    ``float64``                   64-bit binary floating point    ``float``                                  ``float(value)``
    ``decimal32``                 32-bit decimal floating point   ``decimal.Decimal``                        ``decimal.Decimal(value)`` normalized to IEEE 754 decimal32
    ``decimal64``                 64-bit decimal floating point   ``decimal.Decimal``                        ``decimal.Decimal(value)`` normalized to IEEE 754 decimal64
    ``decimal128``                128-bit decimal floating point  ``decimal.Decimal``                        ``decimal.Decimal(value)`` normalized to IEEE 754 decimal128
    ``complex32``                 complex with `float32` values   ``complex``                                ``complex(value)`` with real and imaginary values truncated to 32 bits
    ``complex64``                 complex with `float64` values   ``complex``                                ``complex(value)``
    ``timestamp``                 Nanosecond timestamp            :py:class:`~streamsx.spl.types.Timestamp`  -
    ``rstring``                   UTF-8 string                    ``str``                                    ``str(value)``
    ``rstring[N]``                Bounded UTF-8 string            ``str``                                    ``str(value)``
    ``ustring``                   UTF-16 string                   ``str``                                    ``str(value)``
    ``blob``                      Sequence of bytes               ``memoryview``                             -
    ``list<T>``                   List with elements of type `T`  ``list``                                   -
    ``list<T>[N]``                Bounded list                    ``list``                                   -
    ``set<T>``                    Set with elements of type `T`   ``set``                                    -
    ``set<T>[N]``                 Bounded set                     ``set``                                    -
    ``map<K,V>``                  Map with typed keys and values  ``dict``                                   -
    ``map<K,V>[N]``               Bounded map, limted to N pairs  ``dict``                                   -
    ``optional<T>``               Optional value of type `T`      Value of type `T`, or None                 Value of for type ``T``
    ``enum{id [,...]}``           Enumeration                     Not supported                              Not supported
    ``xml``                       XML value                       Not supported                              Not supported
    ``tuple<type name [, ...]>``  Nested tuple                    Not supported                              Not supported
    ============================  ==============================  =========================================  =======================================

    .. note::
        Type `optional<T>` requires IBM Streams 4.3 or later.

    Python representation is how an attribute value in a structured schema is passed into a Python function. 

    Conversion from Python indicates how a value from Python is converted to an attribute value in a structured schema.
    For example a value ``v`` assigned to ``float64`` attribute is converted as though ``float(v)`` is called first,
    thus ``v`` may be a ``float``, ``int`` or any type that has a ``__float__`` method.
    
    When a type is not supported in Python it can only be used in a schema used for streams produced and consumed by invocation of SPL operators.

    A `StreamSchema` can be created by passing a string of the
    form ``tuple<...>`` or by passing the name of an SPL type from
    an SPL toolkit, for example ``com.ibm.streamsx.transportation.vehicle::VehicleLocation``.

    Attribute names must start with an ASCII letter or underscore, followed by ASCII letters, digits, or underscores.

    When a tuple on a structured stream is passed into a Python callable it
    is converted to a ``dict``, ``tuple`` or named tuple object containing all attributes of the stream tuple.
    See :py:meth:`style`, :py:meth:`as_dict` and :py:meth:`as_tuple` for details.

    When a Python object is submitted to a structured stream,
    for example as the return from the function invoked in a 
    :py:meth:`~streamsx.topology.topology.Stream.map` with the
    `schema` parameter set, it must be:

         * A Python ``dict``. Attributes are set by name using value in the dict for the name. If a value does not exist (the name does not exist as a key) or is set to `None` then the attribute has its default value, zero, false, empty list or string etc.
         * A Python ``tuple`` or named tuple. Attributes are set by position, with the first attribute being the value at index 0 in the Python `tuple`. If a value does not exist (the tuple has less values than the structured schema) or is set to `None` then the attribute has its default value, zero, false, empty list or string etc.

    Args:
        schema(str): Schema definition. Either a schema definition or the name of an SPL type.
    """
    def __init__(self, schema):
        schema = schema.strip()
        self._spl_type = not schema.startswith("tuple<")
        self._schema=schema
        if not self._spl_type:
            parser = _SchemaParser(schema)
            self._types = parser._parse()

        self._style = self._default_style()
            
    def _set(self, schema):
        """Set a schema from another schema"""
        if isinstance(schema, CommonSchema):
            self._spl_type = False
            self._schema = schema.schema()
            self._style = self._default_style()
        else:
            self._spl_type = schema._spl_type
            self._schema = schema._schema
            self._style = schema._style

    @property
    def style(self):
        """Style stream tuples will be passed into a callable.

        For the common schemas the style is fixed:

            * ``CommonSchema.Python`` - ``object`` - Stream tuples are arbitrary objects.
            * ``CommonSchema.String`` - ``str`` - Stream tuples are unicode strings.
            * ``CommonSchema.Json`` - ``dict`` - Stream tuples are a ``dict`` that represents the JSON object.

        For a structured schema the supported styles are:

            * ``dict`` - Stream tuples are passed as a ``dict`` with the key being the attribute name and and the value the attribute value. This is the default.

                * E.g. with a schema of ``tuple<rstring id, float32 value>`` a value is passed as ``{'id':'TempSensor', 'value':20.3}``.

            * ``tuple`` - Stream tuples are passed as a ``tuple`` with the value being the attributes value in order. A schema is set to pass stream tuples as tuples using :py:meth:`as_tuple`.

                * E.g. with a schema of ``tuple<rstring id, float32 value>`` a value is passed as ``('TempSensor', 20.3)``.

            * ``namedtuple`` - Stream tuples are passed as a named tuple (see ``collections.namedtuple``) with the value being the attributes value in order. Field names correspond to the attribute names of the schema. A schema is set to pass stream tuples as named tuples using :py:meth:`as_tuple` setting the `named` parameter.

        Returns:
            type: Class of tuples that will be passed into callables.

        .. versionadded:: 1.8
        .. versionadded:: 1.9 Support for namedtuple.
        """
        return self._style

    def _default_style(self):
        if self._spl_type:
            return _spl_dict
        return _SCHEMA_COMMON_STYLES[self.schema()] if is_common(self) else _spl_dict

    def _copy(self, style=None):
        if style is None:
            return self
        if self._style is style:
            return self
        # Cannot change style of common schemas
        if is_common(self):
            return self
        c = StreamSchema(self.schema())
        c._style = style
        return c

    def _make_named_tuple(self, name):
        if self._spl_type:
            return tuple
        if name is True:
            name = 'StreamTuple'
        nt = typing.NamedTuple(name, _attribute_pytypes(self._types))

        nt._splpy_namedtuple = name
        return nt

    def as_tuple(self, named=None):
        """
        Create a structured schema that will pass stream tuples into callables as ``tuple`` instances.

        If this instance represents a common schema then it will be returned
        without modification. Stream tuples with common schemas are
        always passed according to their definition.

        **Passing as tuple**

        When `named` evaluates to ``False`` then each stream tuple will
        be passed as a ``tuple``. For example with a structured schema
        of ``tuple<rstring id, float64 value>`` a value is passed as
        ``('TempSensor', 27.4)`` and access to the first attribute
        is ``t[0]`` and the second as ``t[1]`` where ``t`` represents
        the passed value..

        **Passing as named tuple**

        When `named` is ``True`` or a ``str`` then each stream tuple will
        be passed as a named tuple. For example with a structured schema
        of ``tuple<rstring id, float64 value>`` a value is passed as
        ``('TempSensor', 27.4)`` and access to the first attribute
        is ``t.id`` (or ``t[0]``) and the second as ``t.value`` (``t[1]``)
        where ``t`` represents the passed value.

        .. warning:: If an schema's attribute name is not a valid Python identifier or
            starts with an underscore then it will be renamed as positional name ``_n``.
            For example, with the schema ``tuple<int32 a, int32 def, int32 id>`` the
            field names are ``a``, ``_1``, ``_2``.


        The value of `named` is used as the name of the named tuple
        class with ``StreamTuple`` used when `named` is ``True``.

        It is not guaranteed that the class of the namedtuple is the
        same for all callables processing tuples with the same
        structured schema, only that the tuple is a named tuple
        with the correct field names.

        Args:
            named: Pass stream tuples as a named tuple.
                If not set then stream tuples are passed as
                instances of ``tuple``.

        Returns:
            StreamSchema: Schema passing stream tuples as ``tuple`` if allowed.

        .. versionadded:: 1.8
        .. versionadded:: 1.9 Addition of `named` parameter.
        """
        if not named:
            return self._copy(tuple)

        if named == True or isinstance(named, str):
            return self._copy(self._make_named_tuple(name=named))

        return self._copy(tuple)

    def as_dict(self):
        """
        Create a structured schema that will pass stream tuples into callables as ``dict`` instances.
        This allows a return to the default calling style for a structured schema.

        If this instance represents a common schema then it will be returned
        without modification. Stream tuples with common schemas are always passed according
        to their definition.

        Returns:
            StreamSchema: Schema passing stream tuples as ``dict`` if allowed.

        .. versionadded:: 1.8
        """
        return self._copy(_spl_dict)

    def schema(self):
        """Private method. May be removed at any time."""
        return self._schema

    def __str__(self):
        """Private method. May be removed at any time."""
        return self._schema

    def spl_json(self):
        _splj = {}
        _splj["type"] = 'spltype'
        _splj["value"] = self.schema()
        return _splj

    def extend(self, schema):
        """
        Extend a structured schema by another.

        For example extending ``tuple<rstring id, timestamp ts, float64 value>``
        with ``tuple<float32 score>`` results in ``tuple<rstring id, timestamp ts, float64 value, float32 score>``.

        Args:
            schema(StreamSchema): Schema to extend this schema by.

        Returns:
            StreamSchema: New schema that is an extension of this schema.
        """
        if self._spl_type:
           raise TypeError("Not supported for declared SPL types")
        base = self.schema()
        extends = schema.schema()
        new_schema = base[:-1] + ',' + extends[6:]
        return StreamSchema(new_schema)

    def __hash__(self):
        return hash(self.schema())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.schema() == other.schema():
                if self.style == other.style:
                    return True
                if _is_typedtuple(self.style) and _is_typedtuple(other.style):
                    return self.style._fields == other.style._fields and self.style._field_types == other.style._field_types
 
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def _fnop_style(schema, op, name):
        """Set an operator's parameter representing the style of this schema."""
        if is_common(schema):
            if name in op.params:
                del op.params[name]
            return
        if _is_pending(schema):
            ntp = 'pending'
        elif schema.style is tuple:
            ntp = 'tuple'
        elif schema.style is _spl_dict:
            ntp = 'dict'
        elif _is_namedtuple(schema.style) and hasattr(schema.style, '_splpy_namedtuple'):
            ntp = 'namedtuple:' + schema.style._splpy_namedtuple
        else:
            return
        op.params[name] = ntp


@enum.unique
class CommonSchema(enum.Enum):
    """
    Common stream schemas for interoperability within Streams applications.

    Streams application can publish streams that are subscribed to by other applications.
    Use of common schemas allow streams connections regardless of the application implementation language.

    Python applications publish streams using :py:meth:`~streamsx.topology.topology.Stream.publish`
    and subscribe using :py:meth:`~streamsx.topology.topology.Topology.subscribe`.
    
     * :py:const:`Python` - Stream constains Python objects.
     * :py:const:`Json` - Stream contains JSON objects.
     * :py:const:`String` - Stream contains strings.
     * :py:const:`Binary` - Stream contains binary tuples.
     * :py:const:`XML` - Stream contains XML documents.
    """
    Python = StreamSchema(_SCHEMA_PYTHON_OBJECT)
    """
    Stream where each tuple is a Python object. Each object
    must be picklable to allow execution in a distributed
    environment where streams can connect processes
    running on the same or different resources.

    Python streams can only be used by Python applications.
    """
    Json = StreamSchema(_SCHEMA_JSON)
    """
    Stream where each tuple is logically a JSON object.

    `Json` can be used as a natural interchange format between Streams applications
    implemented in different programming languages. All languages supported by
    Streams support publishing and subscribing to JSON streams.

    A Python callable receives each tuple as a `dict` as though it was
    created from ``json.loads(json_formatted_str)`` where `json_formatted_str`
    is the JSON formatted representation of tuple.

    Python objects that are to be converted to JSON objects
    must be supported by `JSONEncoder`. If the object is not a `dict`
    then it will be converted to a JSON object with a single key `payload`
    containing the value.
    """
    String = StreamSchema(_SCHEMA_STRING)
    """
    Stream where each tuple is a string.

    `String` can be used as a natural interchange format between Streams applications
    implemented in different programming languages. All languages supported by
    Streams support publishing and subscribing to string streams.

    A Python callable receives each tuple as a `str` object.

    Python objects are converted to strings using ``str(obj)``.
    """
    Binary = StreamSchema(_SCHEMA_BINARY)
    """
    Stream where each tuple is a binary object (sequence of bytes).

    .. warning:: `Binary` is not yet supported for Python applications.
    """
    XML = StreamSchema(_SCHEMA_XML)
    """
    Stream where each tuple is an XML document.

    .. warning:: `XML` is not yet supported for Python applications.
    """

    def schema(self):
        """Private method. May be removed at any time."""
        return self.value.schema()

    def spl_json(self):
        return self.value.spl_json()

    def extend(self, schema):
        """Extend a structured schema by another.

        Args:
            schema(StreamSchema): Schema to extend this schema by.

        Returns:
            StreamSchema: New schema that is an extension of this schema.
        """
        return self.value.extend(schema)

    def __str__(self):
        return str(self.schema())

def _from_named_tuple(nt):
    spl_types = []
    for name in nt._fields:
        spl_types.append(_spl_from_type(nt._field_types[name]))

    td = 'tuple<'
    for i in range(len(nt._fields)):
        if i:
            td += ', '
        td += spl_types[i]
        td += ' '
        td += nt._fields[i]
    td += '>'
    return StreamSchema(td).as_tuple(named=nt.__name__)

def _spl_from_type(type_):
    _init_type_mappings()
    if type_ in _PYTYPE_TO_SPL:
        return _PYTYPE_TO_SPL[type_]

    # See https://bugs.python.org/issue34568
    # isinstance,issubclass no longer work in Python 3.7
    if hasattr(type_, '__origin__') and hasattr(type_, '__args__'):
        if len(type_.__args__) == 1:
            et = type_.__args__[0]
            if typing.List[et] == type_:
                return 'list<' + _spl_from_type(et) + '>'
            if typing.Set[et] == type_:
                return 'set<' + _spl_from_type(et) + '>'
        elif len(type_.__args__) == 2:
            kt = type_.__args__[0]
            vt = type_.__args__[1]
            if typing.Mapping[kt, vt] == type_:
                return 'map<' + _spl_from_type(kt) + ', ' + _spl_from_type(vt) + '>'
    if hasattr(type_, '__args__') and len(type_.__args__) == 2:
        if type(None) in type_.__args__:
            et = type_.__args__[0] if type_.__args__[1] is type(None) else type_.__args__[1]
            if typing.Optional[et] == type_:
                return 'optional<' + _spl_from_type(et) + '>'
    raise ValueError("Unsupported type: " + str(type_))

def _type_from_spl(type_):
    _init_type_mappings()
    if type_ in _SPLTYPE_TO_PY:
        return _SPLTYPE_TO_PY[type_]

    if isinstance(type_, tuple):
        if type_[0] == 'list':
            return typing.List[_type_from_spl(type_[1])]
        if type_[0] == 'set':
            return typing.Set[_type_from_spl(type_[1])]
        if type_[0] == 'map':
            return typing.Mapping[_type_from_spl(type_[1][0]), _type_from_spl(type_[1][1])]
        if type_[0] == 'optional':
            return typing.Optional[_type_from_spl(type_[1])]
    raise ValueError("Unsupported type: " + type_)

_PYTYPE_TO_SPL = {}
_SPLTYPE_TO_PY = {}
def _init_type_mappings():
    global _PYTYPE_TO_SPL
    if not _PYTYPE_TO_SPL:
        import streamsx.spl.types
        _PYTYPE_TO_SPL = {
            str:'rstring', bool:'boolean', int:'int64', float:'float64',
            complex:'complex64', decimal.Decimal:'decimal128',
            streamsx.spl.types.Timestamp:'timestamp',
            datetime.datetime:'timestamp',
            bytes:'blob' }

        global _SPLTYPE_TO_PY
        _SPLTYPE_TO_PY = {
            'rstring': str, 'boolean':bool,
            'int8':int, 'int16':int, 'int32':int, 'int64':int,
            'uint8':int, 'uint16':int, 'uint32':int, 'uint64':int,
            'float32':float, 'float64':float,
            'complex32':complex, 'complex64':complex,
            'decimal32':decimal.Decimal,
            'decimal64':decimal.Decimal,
            'decimal128':decimal.Decimal,
            'timestamp':streamsx.spl.types.Timestamp,
            'blob': bytes}



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2019

import sys
import sysconfig
import os
import argparse
import requests
import warnings
import urllib3
import datetime
import json
import locale
import re

import streamsx.topology.context
from streamsx.rest_primitives import Instance
from streamsx.build import BuildService



###########################################
# submitjob
###########################################
def _submitjob_parser(subparsers):
    job_submit = subparsers.add_parser('submitjob', help='Submit an application bundle')
    job_submit.add_argument('sabfile', help='Location of sab file.', metavar='sab-pathname')
    job_submit.add_argument('--jobConfig', '-g', help='Specifies the name of an external file that defines a job configuration overlay', metavar='file-name')
    job_submit.add_argument('--jobname', help='Specifies the name of the job.', metavar='job-name')
    job_submit.add_argument('--jobgroup', '-J', help='Specifies the job group', metavar='jobgroup-name')
    job_submit.add_argument('--outfile', help='Specifies the path and file name of the output file in which the command writes the list of submitted job IDs', metavar='file-name')
    job_submit.add_argument('--P', '-P', help='Specifies a submission-time parameter and value for the job', action='append', metavar='parameter-name')
    _user_arg(job_submit)

def _submitjob(instance, cmd_args, rc):
    """Submit a job."""
    job_config = None

    if cmd_args.jobConfig:
        with open(cmd_args.jobConfig) as fd:
            job_config = streamsx.topology.context.JobConfig.from_overlays(json.load(fd))
    else:
        job_config = streamsx.topology.context.JobConfig()

    if cmd_args.jobname:
        job_config.job_name = cmd_args.jobname

    if cmd_args.jobgroup:
        # Will throw a 500 internal error on instance.submit_job if jobgroup doesn't already exist
        job_config.job_group = cmd_args.jobgroup

    if cmd_args.P:
        for param in cmd_args.P:
            name_value_pair = param.split("=")
            if len(name_value_pair) != 2:
                raise ValueError("The format of the following submission-time parameter is not valid: {}. The correct syntax is: <name>=<value>".format(param))
            else:
                job_config.submission_parameters[name_value_pair[0]] = name_value_pair[1]

    job = instance.submit_job(bundle=cmd_args.sabfile, job_config=job_config)

    if job:
        print("The following number of applications were submitted to the {} instance: {}.".format(instance.id, 1))
        print("Submitted job IDs: {}".format(job.id), file=sys.stderr)
    else:
        raise Exception("Error in creating Job")

    # If --outfile, write jobID to file
    if cmd_args.outfile:
        with open(cmd_args.outfile, 'w') as my_file:
            my_file.write(str(job.id) + '\n')

    return (rc, job)

###########################################
# canceljob
###########################################
def _canceljob_parser(subparsers):
    job_cancel = subparsers.add_parser('canceljob', help='Cancel a job.')
    job_cancel.add_argument('--force', action='store_true', help='Stop the service even if jobs are running.', default=False)
    job_cancel.add_argument('--collectlogs', help='Specifies to collect the log and trace files for each processing element that is associated with the job', action='store_true')
    job_cancel.add_argument('jobid', help='Specifies a list of job IDs.', nargs='*', metavar='jobid')
    # Only 1 of these arguments --jobs, --jobnames, --file can be specified at any given time when running this command
    g1 = job_cancel.add_argument_group(title='jobs jobnames file group', description='One of these options must be chosen.')
    group = g1.add_mutually_exclusive_group(required=False)
    group.add_argument('--jobs', '-j', help='Specifies a list of job IDs.', metavar='job-id')
    group.add_argument('--jobnames', help='Specifies a list of job names', metavar='job-names')
    group.add_argument('--file', '-f', help='Specifies the file that contains a list of job IDs, one per line', metavar='file-name')

    _user_arg(job_cancel)

def _canceljob(instance, cmd_args, rc):
    """Cancel a job."""
    job_ids_to_cancel = []
    job_names_to_cancel = []

    # Check for mutually exclusive arguments, just in case argparse doesn't catch it
    temp = len(list(filter(None, [cmd_args.jobid, cmd_args.jobs, cmd_args.jobnames, cmd_args.file])))
    if temp > 1:
        raise ValueError("Arguments jobid, --jobs, --jobnames, --file are mutually exclusive")

    # get list of job IDs to cancel
    if cmd_args.jobid:
        for x in cmd_args.jobid:
            job_ids = x.split(',')
            job_ids = [job.strip() for job in job_ids]
            job_ids_to_cancel.extend(job_ids)

    # if --jobs, get list of job IDs to cancel
    if cmd_args.jobs:
        job_ids = cmd_args.jobs.split(',')
        job_ids_to_cancel.extend(job_ids)

    # if --jobnames, get list of jobnames to cancel
    if cmd_args.jobnames:
        job_names = cmd_args.jobnames.split(',')
        job_names_to_cancel.extend(job_names)

    # if --file, get list of job IDs to cancel from file
    if cmd_args.file:
        with open(cmd_args.file, 'r') as my_file:
            job_ids = [line.rstrip() for line in my_file if not line.isspace()]
            job_ids_to_cancel.extend(job_ids)

    # Filter out any blank/empty strings in job_ids_to_cancel and job_names_to_cancel
    # Ex. 'canceljob job1.id , job2.id' -> job_ids_to_cancel = ['job1.id', '', '', 'job2.id'], should be = ['job1.id', 'job2.id']
    job_ids_to_cancel = list(filter(None, job_ids_to_cancel))
    job_names_to_cancel = list(filter(None, job_names_to_cancel))

    # If no jobs to cancel, raise error
    if not job_ids_to_cancel and not job_names_to_cancel:
        raise Exception("No jobs provided")

    # Check if job w/ job ID exists, and if so cancel it
    for x in job_ids_to_cancel:
        if x.isnumeric():
            try:
                job = instance.get_job(id=str(x))
                _job_cancel(instance, x, cmd_args.collectlogs, cmd_args.force)
            except:
                print("The following job ID was not found {}".format(x), file=sys.stderr)
                print("The following job ID cannot be canceled: {}. See the previous error message".format(x), file=sys.stderr)
                rc = 1
        else:
            raise ValueError("The following job identifier is not valid: {}. Specify a job identifier that is numeric and try the request again.".format(x))

    # Check if job w/ job name exists, and if so cancel it
    for x in job_names_to_cancel:
        # if jobname contains a space, its invalid
        if ' ' in x:
            print("{} is not a valid job name. Either its size is longer than 1024 characters or it includes some invalid characters.".format(x), file=sys.stderr)
            rc = 1
            continue
        jobs = instance.get_jobs(name=str(x))
        if jobs:
            job = jobs[0]
            _job_cancel(instance, job.id, cmd_args.collectlogs, cmd_args.force)
        else:
            print("The following job name is not found: {}. Specify a job name that is valid and try the request again.".format(x), file=sys.stderr)
            rc = 1

    return (rc, None)

def _job_cancel(instance, job_id=None, collectlogs=False, force=False):
    job = instance.get_job(id=str(job_id))
    if collectlogs:
        log_path = job.retrieve_log_trace()
        if log_path:
            print("The log files for the {} job ID will be collected in the following files: {}".format(job_id, log_path))
        else:
            raise Exception("Retrieval of job's logs is not supported in this version of IBM Streams")

    # Cancel job, and check if successful
    val = job.cancel(force)
    if val:
        print("The following job ID was canceled: {}. The job was in the {} instance.".format(job_id, instance.id))
    else:
        raise Exception("One or more jobs failed to stop")


###########################################
# lsjobs
###########################################
def _lsjobs_parser(subparsers):
    job_ls = subparsers.add_parser('lsjobs', help='List the jobs for a given instance')
    job_ls.add_argument('--jobs', '-j', help='Specifies a list of job IDs.', metavar='job-id')
    job_ls.add_argument('--users', '-u', help='Specifies to select from this list of user IDs', metavar='user')
    job_ls.add_argument('--jobnames', help='Specifies a list of job names', metavar='job-names')
    job_ls.add_argument('--fmt', help='Specifies the presentation format', default='%Tf', metavar='format-spec')
    job_ls.add_argument('--xheaders', help='Specifies to exclude headings from the report', action='store_true')
    job_ls.add_argument('--long', '-l', help='Reports launch count, full host names, and all of the operator instance names for the PEs.', action='store_true')
    job_ls.add_argument('--showtimestamp', help='Specifies to show a time stamp in the output to indicate when the command was run.', action='store_true')

    _user_arg(job_ls)

def _lsjobs(instance, cmd_args, rc):
    """view jobs"""
    jobs = instance.get_jobs()
    instance_id = instance.id

    # Omit header if xheaders is present
    if cmd_args.xheaders:
        cmd_args.showtimestamp = False
        instance_id = None

    # If --users argument (ie given list of user ID's), filter jobs by these user ID's
    if (cmd_args.users):
        users = cmd_args.users.split(",")
        jobs = [job for job in jobs if job.startedBy in users]

    # If --jobs argument (ie given list of job ID's), filter jobs by these ID's
    if (cmd_args.jobs):
        job_ids = cmd_args.jobs.split(",")
        jobs = [job for job in jobs if job.id in job_ids]

    # If --jobsnames argument (ie given list of job names), filter jobs by these user job names
    if (cmd_args.jobnames):
        job_names = cmd_args.jobnames.split(",")
        jobs = [job for job in jobs if job.name in job_names]

    # If --showtimestamp, give current date
    time_stamp = None
    LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
    if cmd_args.showtimestamp:
        time_stamp = datetime.datetime.now().replace(microsecond=0).replace(tzinfo=LOCAL_TIMEZONE).isoformat()

    headers = ["Id", "State", "Healthy", "User", "Date", "Name", "Group"]

    # If --long option, add productVersion to the headers and get the corresponding value for each job
    if cmd_args.long:
        headers.append('ProductVersion')

    # pre process the data so output is formatted nicely
    h_length = [len(x) for x in headers] # header_length
    job_data = []
    for job in jobs:
        jobHealth = "yes" if job.health == "healthy" else "no"
        jobTime = datetime.datetime.fromtimestamp(job.submitTime/1000, tz = LOCAL_TIMEZONE).isoformat() # job.submitTime/1000 to convert ms to sec
        jobGroup = job.jobGroup.split("/")[-1]
        data = [job.id, job.status.capitalize(), jobHealth, job.startedBy, jobTime, job.name, jobGroup]

        h_length[0] = max(len(job.id), h_length[0])
        h_length[1] = max(len(job.status), h_length[1])
        h_length[2] = max(len(jobHealth), h_length[2])
        h_length[3] = max(len(job.startedBy), h_length[3])
        h_length[4] = max(len(jobTime), h_length[4])
        h_length[5] = max(len(job.name), h_length[5])
        h_length[6] = max(len(jobGroup), h_length[6])

        # If --long option, add productVersion to the job data
        if cmd_args.long:
            prod_version = job.productVersion
            h_length[7] = max(len(prod_version), h_length[7])
            data.append(prod_version)
        job_data.append(data)

    # Output the data

    # If default output format
    if cmd_args.fmt == '%Tf':
        if time_stamp:
            print("Date: " + time_stamp)

        if instance_id:
            print("Instance: " + instance_id)
            # String that represents the header w/ formatting for nice output
            header_string = '{:>{id_w}}  {:<{state_w}}  {:<{health_w}}  {:<{user_w}}  {:<{date_w}}  {:<{name_w}}  {:<{group_w}}'.format(
                "Id", "State", "Healthy", "User", "Date", "Name", "Group",
                id_w=h_length[0], state_w=h_length[1],
                health_w=h_length[2], user_w=h_length[3],
                date_w=h_length[4], name_w=h_length[5], group_w=h_length[6])

            # If --long option, we also want the header to display ProductVersion
            if cmd_args.long:
                header_string += '  {:<{prod_w}}'.format("ProductVersion", prod_w=h_length[7])
            print(header_string)

        for job in job_data:
            # String that represents a job w/ formatting for nice output
            job_string = '{:>{id_w}}  {:<{state_w}}  {:<{health_w}}  {:<{user_w}}  {:<{date_w}}  {:<{name_w}}  {:<{group_w}}'.format(
                job[0], job[1], job[2], job[3], job[4], job[5], job[6],
                id_w=h_length[0], state_w=h_length[1],
                health_w=h_length[2], user_w=h_length[3],
                date_w=h_length[4], name_w=h_length[5], group_w=h_length[6])

            # If --long option, we also want the job to display ProductVersion
            if cmd_args.long:
                job_string += '  {:<{prod_w}}'.format(job[7], prod_w=h_length[7])
            print(job_string)

    elif cmd_args.fmt == "%Mf":
        border = '=' * 50

        if instance_id:
            print(border)
            if time_stamp:
                print("Date: " + time_stamp)
            print("Instance: " + instance_id)

        w1 = len(max(headers, key=len))
        w2 = max(h_length)
        for job in job_data:
            print(border)
            for header, job_item in zip(headers, job):
                print('{:{w1}}  :  {: <{w2}}'.format(header, job_item, w1=w1, w2=w2))
        print(border)

    elif cmd_args.fmt == "%Nf":
        for job in job_data:
            toPrint = ''
            for header, row in zip(headers, job):
                toPrint += '{}  :  {} '.format(header, row)
            print(toPrint)

    return (rc, None)

###########################################
# appconfig
###########################################

# ls-appconfig
def _lsappconfig_parser(subparsers):
    appconfig_ls = subparsers.add_parser('lsappconfig', help='Retrieve a list of configurations for making a connection to an external application')
    appconfig_ls.add_argument('--fmt', help='Specifies the presentation format', default='%Tf', metavar='format-spec')
    _user_arg(appconfig_ls)

def _lsappconfig(instance, cmd_args, rc):
    """view appconfigs"""
    configs = instance.get_application_configurations()
    locale.setlocale(locale.LC_ALL, '') # Needed to correctly display local date and time format
    LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo # Needed to get timezone

    headers = ["Id", "Owner", "Created", "Modified", "Description"]

    # pre process the data so output is formatted nicely
    h_length = [len(x) for x in headers] # header_length
    config_data = []
    for config in configs:
        createDate = datetime.datetime.fromtimestamp(config.creationTime/1000, tz = LOCAL_TIMEZONE).strftime("%x %X %Z")
        lastModifiedDate = datetime.datetime.fromtimestamp(config.lastModifiedTime/1000, tz = LOCAL_TIMEZONE).strftime("%x %X %Z")

        data = [config.name, config.owner, createDate, lastModifiedDate, config.description]
        config_data.append(data)

        h_length[0] = max(len(config.name), h_length[0])
        h_length[1] = max(len(config.owner), h_length[1])
        h_length[2] = max(len(createDate), h_length[2])
        h_length[3] = max(len(lastModifiedDate), h_length[3])
        h_length[4] = max(len(config.description), h_length[4])

    # Output the data

    # If default output format
    if cmd_args.fmt == '%Tf':
        print('{:>{id_w}}  {:<{owner_w}}  {:<{created_w}}  {:<{mod_w}}  {:<{desc_w}}'.format(
            "Id", "Owner", "Created", "Modified", "Description",
            id_w=h_length[0], owner_w=h_length[1],
            created_w=h_length[2], mod_w=h_length[3],
            desc_w=h_length[4]))

        for config in config_data:
            print('{:>{id_w}}  {:<{owner_w}}  {:<{created_w}}  {:<{mod_w}}  {:<{desc_w}}'.format(
            config[0], config[1], config[2], config[3], config[4],
            id_w=h_length[0], owner_w=h_length[1],
            created_w=h_length[2], mod_w=h_length[3],
            desc_w=h_length[4]))

    elif cmd_args.fmt == "%Mf":
        border = '=' * 50

        w1 = len(max(headers, key=len))
        w2 = max(h_length)
        for config in config_data:
            print(border)
            for header, config_item in zip(headers, config):
                print('{:{w1}}  :  {: <{w2}}'.format(header, config_item, w1=w1, w2=w2))
        print(border)

    elif cmd_args.fmt == "%Nf":
        for config in config_data:
            toPrint = ''
            for header, row in zip(headers, config):
                toPrint += '{}  :  {} '.format(header, row)
            print(toPrint)

    return (rc, None)

# rm-appconfig
def _rmappconfig_parser(subparsers):
    appconfig_rm = subparsers.add_parser('rmappconfig', help='Removes a configuration that is used for making a connection to an external application')
    appconfig_rm.add_argument('config_name', help='Name of the app config', metavar='config-name')
    appconfig_rm.add_argument('--noprompt', help='Specifies to suppress confirmation prompts.', action='store_true')
    _user_arg(appconfig_rm)

def _rmappconfig(instance, cmd_args, rc):
    """remove an appconfig"""
    config_name = cmd_args.config_name
    configs = instance.get_application_configurations(name = config_name)
    if (not configs):
        raise NameError("The {} application configuration does not exist in the {} instance".format(config_name, instance.id))
    app_config = instance.get_application_configurations(name = config_name)[0]

    # No confirmation required, delete
    if cmd_args.noprompt:
        app_config.delete()
        print("The {} application configuration was removed successfully for the {} instance".format(config_name, instance.id))
    else:
        response = input("Do you want to remove the application configuration {} from the {} instance? Enter 'y' to continue or 'n' to cancel: ".format(config_name, instance.id))
        if response == "y":
            app_config.delete()
            print("The {} application configuration was removed successfully for the {} instance".format(config_name, instance.id))

    return (rc, None)

# mk-appconfig
def _mkappconfig_parser(subparsers):
    appconfig_mk = subparsers.add_parser('mkappconfig', help='Creates a configuration that enables connection to an external application')
    appconfig_mk.add_argument('config_name', help='Name of the app config', metavar='config-name')
    appconfig_mk.add_argument('--property', action='append', help='Specifies a property name and value pair to add to or change in the configuration', metavar='name=value')
    appconfig_mk.add_argument('--propfile', help='Specifies the path to a file that contains a list of application configuration properties for connecting to an external application', metavar='property-file')
    appconfig_mk.add_argument('--description', help='Specifies a description for the application configuration', metavar='description')
    _user_arg(appconfig_mk)

def _mkappconfig(instance, cmd_args, rc):
    """create an appconfig"""
    config_name, config_props, config_description = _get_config_details(cmd_args, mk=True)
    appconfig = None

    # Check if config exists by that name, if so don't do anything
    if instance.get_application_configurations(config_name):
        raise Exception("The {} application configuration already exists in the following {} instance".format(config_name, instance.id))
    else:
        # No appconfig exists by that name, create new one
        appconfig =  instance.create_application_configuration(name=config_name, properties=config_props, description=config_description)
        if appconfig:
            print("The {} application configuration was created successfully for the {} instance".format(config_name, instance.id))
        else:
            # Failed to create appconfig
            rc = 1

    return (rc, appconfig)

# ch-appconfig
def _chappconfig_parser(subparsers):
    appconfig_ch = subparsers.add_parser('chappconfig', help='Change the configuration properties that are used to make a connection to an external application')
    appconfig_ch.add_argument('config_name', help='Name of the app config', metavar='config-name')
    appconfig_ch.add_argument('--property', action='append', help='Specifies a property name and value pair to add to or change in the configuration', metavar='name=value')
    appconfig_ch.add_argument('--description', help='Specifies a description for the application configuration', metavar='description')
    _user_arg(appconfig_ch)

def _chappconfig(instance, cmd_args, rc):
    """change an appconfig"""
    config_name, config_props, config_description = _get_config_details(cmd_args, mk=False)
    new_app_config = None
    # Check if config exists by that name, if so update it
    if instance.get_application_configurations(config_name):
        appconfig = instance.get_application_configurations(config_name)[0]
        new_app_config = appconfig.update(properties=config_props, description=config_description)
        if new_app_config:
            print("The {} application configuration was updated successfully for the {} instance".format(config_name, instance.id))
        else:
            rc = 1
    else:
        # No appconfig exists by that name
        raise NameError("The {} application configuration does not exist in the {} instance".format(config_name, instance.id))

    return (rc, new_app_config)


def _get_config_details(cmd_args, mk):
    """A helper function for mkappconfig and chappconfig that gets the name, properties and description of an appconfig
    Arguments:
        cmd_args {List} -- The args passed in the command line
        mk {Boolean} -- Describes whether or not to check for a propFile arg (only mkappconfig can have a --propfile arg)

    Returns:
        config_name {String} -- The name of the appconfig
        config_props {Dictionary} -- key-value pair properties of the appconfig
        config_description {String} -- description of the appconfig
    """
    config_name = cmd_args.config_name
    config_props = {}
    config_description = None
    # If mkappconfig command (bc only this command allows for a propFile) Get props_file if given
    if mk:
        if cmd_args.propfile:
            my_file = open(cmd_args.propfile)
            prop_list = [line.rstrip() for line in my_file if not line.isspace()]
            my_file.close()
            config_props = _create_appconfig_props(config_props, prop_list)

    # Get props from command line
    # Name's specified via command line (ie --property option) override names specified in the prop file
    if cmd_args.property:
        prop_list = cmd_args.property
        config_props = _create_appconfig_props(config_props, prop_list)

    # Create the config description
    if cmd_args.description:
        config_description = cmd_args.description

    return config_name, config_props, config_description

def _create_appconfig_props(config_props, prop_list):
    """A helper function for mk-appconfig and ch-appconfig that creates the config properties dictionary

    Arguments:
        config_props {Dictionary} -- A dictionary with key-value pairs that gives the properties in an appconfig
        prop_list {List} -- A list with items of the form <Key>=<Value> that describe a given property in an appconfig

    Returns:
        config_props {Dictionary} -- A dictionary with key-value pairs that gives the properties in an appconfig
    """
    # Iterate through list of properties, check if correct format, convert each to name/value pairs add to config_props dict
    # Ex. good proplist is ['name1=value1', 'name2=value2']
    # Ex. bad proplist is ['name1=valu=e1', 'name2=value2']
    for prop in prop_list:
        name_value_pair = prop.split("=")
        if (len(name_value_pair) != 2):
            raise ValueError("The format of the following property specification is not valid: {}. The correct syntax is: <name>=<value>".format(prop))
        config_props[name_value_pair[0]] = name_value_pair[1]
    return config_props

# get-appconfig
def _getappconfig_parser(subparsers):
    appconfig_get = subparsers.add_parser('getappconfig', help='Displays the properties of a configuration that enables connection to an external application')
    appconfig_get.add_argument('config_name', help='Name of the app config', metavar='config-name')
    _user_arg(appconfig_get)

def _getappconfig(instance, cmd_args, rc):
    """get an appconfig"""
    config_name = cmd_args.config_name
    configs = instance.get_application_configurations(name = config_name)
    # Check if any configs by that name
    if (not configs):
        raise NameError("No application configuration by the name {}".format(config_name))
    config = configs[0]
    config_props = config.properties
    if not config_props:
        raise Exception("The {} application configuration has no properties defined".format(config_name))

    for key, value in config_props.items():
        try:
            json_value = json.loads(value)
            json_value = json.dumps(json_value, indent=2)
        except ValueError:
            json_value = value
        print(key + "=" + json_value)
    return (rc, config_props)

###########################################
# rmtoolkit
###########################################
def _rmtoolkit_parser(subparsers):
    toolkit_rm = subparsers.add_parser('rmtoolkit', help='Remove toolkits from a build service')
    g1 = toolkit_rm.add_argument_group(title='Toolkit selection', description='Selects which toolkits will be removed.')
    group = g1.add_mutually_exclusive_group(required=True)
    group.add_argument('--id', '-i', help='Specifies the id of the toolkit to remove', metavar='toolkit-id')
    group.add_argument('--name', '-n', help='Remove all toolkits with this name', metavar='toolkit-name')
    group.add_argument('--regex', '-r', help='Remove all toolkits where the name matches the given regex pattern', metavar='toolkit-regex')
    _user_arg(toolkit_rm)

def _rmtoolkit(instance, cmd_args, rc):
    # Get all toolkits from the build_server
    build_server = BuildService.of_endpoint(verify=False if cmd_args.disable_ssl_verify else None)

    tk_to_delete = []
    return_message = None

    # Find the toolkit matching toolkitid
    if cmd_args.id:
        try:
            matching_toolkit = build_server.get_toolkit(cmd_args.id)
            tk_to_delete.append(matching_toolkit)
        except ValueError:
            pass

    # Find all toolkits with toolkitname
    elif cmd_args.name:
        remote_toolkits = build_server.get_toolkits()
        matching_toolkits = [x for x in remote_toolkits if x.name == cmd_args.name]
        if matching_toolkits:
            tk_to_delete.extend(matching_toolkits)

    # Find all toolkits where the name matches toolkitregex
    elif cmd_args.regex:
        matching_toolkits = build_server.get_toolkits(name=cmd_args.regex)
        if matching_toolkits:
            tk_to_delete.extend(matching_toolkits)

    # If there are any toolkits to delete, delete them
    for tk in tk_to_delete:
        val = tk.delete()
        if not val:
            # If tk fails to delete, set error code and message
            rc = 1
            return_message = '1 or more toolkits failed to delete'

    return (rc, return_message)

###########################################
# lstoolkit
###########################################
def _lstoolkit_parser(subparsers):
    toolkit_ls = subparsers.add_parser('lstoolkit', help='List toolkits on the build service')
    g1 = toolkit_ls.add_argument_group(title='Toolkit selection', description='Selects which toolkits will be listed.')
    group = g1.add_mutually_exclusive_group(required=True)
    group.add_argument('--all', '-a', action='store_true', help='List all toolkits')
    group.add_argument('--id', '-i', help='Specifies the id of the toolkit to list', metavar='toolkit-id')
    group.add_argument('--name', '-n', help='List all toolkits with this name', metavar='toolkit-name')
    group.add_argument('--regex', '-r', help='List all toolkits where the name matches the given regex pattern', metavar='toolkit-regex')
    _user_arg(toolkit_ls)

def _lstoolkit(instance, cmd_args, rc):
    # Get all toolkits from the build_service
    build_server = BuildService.of_endpoint(verify=False if cmd_args.disable_ssl_verify else None)

    return_message = None

    # Find the toolkit matching toolkitid
    if cmd_args.id:
        try:
            matching_toolkit = build_server.get_toolkit(cmd_args.id)
            tk_to_list = [matching_toolkit]
        except ValueError:
            tk_to_list = []

    # Find all toolkits with toolkitname
    elif cmd_args.name:
        remote_toolkits = build_server.get_toolkits()
        tk_to_list = [x for x in remote_toolkits if x.name == cmd_args.name]

    # Find all toolkits where the name matches toolkitregex
    elif cmd_args.regex:
        tk_to_list = build_server.get_toolkits(name=cmd_args.regex)
    elif cmd_args.all:
        tk_to_list = build_server.get_toolkits()

    tk_to_list.sort(key=lambda tk : (tk.name, tk.version))

    headers = ["Name", "Version", 'RequiredProductVersion']

    # pre process the data so output is formatted nicely
    h_length = [len(x) for x in headers] # header_length
    for tk in tk_to_list:
        h_length[0] = max(len(tk.name), h_length[0])
        h_length[1] = max(len(tk.version), h_length[1])
        h_length[2] = max(len(tk.requiredProductVersion), h_length[2])

    fmt = '{:<{name_w}}  {:<{ver_w}}  {:<{pv_w}}'

    print(fmt.format(headers[0], headers[1], headers[2],
        name_w=h_length[0], ver_w=h_length[1], pv_w=h_length[2]))
    for tk in tk_to_list:
         print(fmt.format(tk.name, tk.version, tk.requiredProductVersion,
             name_w=h_length[0], ver_w=h_length[1], pv_w=h_length[2]))

    return (rc, return_message)

###########################################
# uploadtoolkit
###########################################
def _uploadtoolkit_parser(subparsers):
    toolkit_up = subparsers.add_parser('uploadtoolkit', help='Upload a toolkit to a build service')
    toolkit_up.add_argument('--path', '-p', help='Path to the toolkit to be uploaded', required=True)
    _user_arg(toolkit_up)

def _uploadtoolkit(instance, cmd_args, rc):
    bs = BuildService.of_endpoint(verify=False if cmd_args.disable_ssl_verify else None)

    return_message = None

    tk = bs.upload_toolkit(cmd_args.path)
    if tk is not None:
        print('Toolkit with id {} uploaded from {}'.format(tk.id, cmd_args.path))
    else:
        return_message = 'Toolkit not uploaded from {}'.format(cmd_args.path)
        print(return_message, file=sys.stderr )
        rc = 1

    return (rc, return_message)

###########################################
# updateoperators
###########################################
def _updateops_parser(subparsers):
    update_ops = subparsers.add_parser('updateoperators', help='Adjust a job configuration while the job is running')
    g1 = update_ops.add_argument_group(title='Job selection', description='One of these options must be chosen.')
    group = g1.add_mutually_exclusive_group(required=True)
    group.add_argument('jobid', help='Specifies a job ID.', nargs='?', metavar='jobid')
    group.add_argument('--jobname', help='Specifies the name of the job.', metavar='job-name')
    update_ops.add_argument('--jobConfig', '-g', help='Specifies the name of an external file that defines a job configuration overlay', metavar='file-name')
    update_ops.add_argument('--parallelRegionWidth', help='Specifies a parallel region name and its width', metavar='parallelRegionName=width')
    update_ops.add_argument('--force', action='store_true', help='Specifies whether to automatically stop the PEs that need to be stopped', default=False)

    _user_arg(update_ops)

def _updateops(instance, cmd_args, rc):
    return_message = None
    job_config_json = None
    job = None

    # Check that job w/ jobID or jobname exists
    if cmd_args.jobid:
        job = instance.get_job(id=str(cmd_args.jobid))
    elif cmd_args.jobname:
        jobs = instance.get_jobs(name=str(cmd_args.jobname))
        if jobs:
            job = jobs[0]

    # job doesn't exist, throw error
    if not job:
        return (1, "The job was not found")

    if cmd_args.jobConfig:
        with open(cmd_args.jobConfig) as fd:
            job_config_json = json.load(fd)
        # If empty JCO passed in, throw error
        if not job_config_json:
            return (1, 'JCO is empty')
    elif cmd_args.parallelRegionWidth:
        # If no JCO, but parallelRegionWidth arg passed in, create empty JCO and later populate w/ parallelRegionWidth
        job_config_json = {}
    else:
        # No JCO or parallelRegionWidth arg, throw error
        return (1, 'A JCO or parallelRegionWidth is required')


    if cmd_args.parallelRegionWidth:
        # Overrides the targetParallelRegion if already present in the JCO
        # else, populate empty JCO w/ parallelRegionWidth
        arr = cmd_args.parallelRegionWidth.split('=')
        if len(arr) != 2:
            raise ValueError("The format of the following submission-time parameter is not valid: {}. The correct syntax is: <name>=<value>".format(arr))
        name, width = arr[0], arr[1]

        entry = {'targetParallelRegion': {'regionName': name, 'newWidth': int(width)}}

        # If JCO is empty, and parallelRegionWidth arg is present, create JCO with arg
        if not job_config_json:
            job_config_json = {
                'jobConfigOverlays' : [
                    {'configInstructions' : {'adjustmentSection': [entry]}
                    }
                ]
            }
        else:
            # Non-empty JCO, and parallelRegionWidth arg is present, override existing in JCO

            # jobConfigOverlays is an array, where only the first jobConfigOverlay is supported
            JCO = job_config_json['jobConfigOverlays'][0]
            # Check if configInstructions already exists
            if 'configInstructions' in JCO:
                cfg_inst = JCO['configInstructions']
                # Overwrite adjustmentSection, since only 1 parallelRegion can be specified
                cfg_inst['adjustmentSection'] = [entry]
            else:
                JCO['configInstructions'] = {'adjustmentSection': [entry]}

    # If --force present, force PE to stop
    if cmd_args.force:
        JCO = job_config_json['jobConfigOverlays'][0]
        JCO['operationConfig'] = {"forcePeStopped": True}

    job_config = streamsx.topology.context.JobConfig.from_overlays(job_config_json)
    json_result = job.update_operators(job_config)

    # --- 1/13/20 JSON result is incorrect until 1Q20 fix ---
    # if json_result:
    #     file_name = str(job.name) + '_' + str(job.id) + '_config.json'
    #     with open(file_name, 'w') as outfile:
    #         json.dump(json_result, outfile)

    if json_result !=0:
        return (1, 'Update operators failed')
    else:
        print('Update operators was started on the {} instance.'.format(instance.id))
        # print('The operator configuration results were written to the following file: {}'.format(file_name))

    return (rc, return_message)

def run_cmd(args=None):
    cmd_args = _parse_args(args)

    if cmd_args.disable_ssl_verify:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    instance = Instance.of_endpoint(
        username=cmd_args.User,
        verify=False if cmd_args.disable_ssl_verify else None)

    switch = {
    "submitjob": _submitjob,
    "canceljob": _canceljob,
    "lsjobs": _lsjobs,
    "lsappconfig": _lsappconfig,
    "rmappconfig":_rmappconfig,
    "mkappconfig": _mkappconfig,
    "chappconfig": _chappconfig,
    "getappconfig": _getappconfig,
    "rmtoolkit": _rmtoolkit,
    "lstoolkit": _lstoolkit,
    "uploadtoolkit": _uploadtoolkit,
    "updateoperators": _updateops,
    }

    extra_info = None
    rc = 0
    try:
        rc, extra_info = switch[cmd_args.subcmd](instance, cmd_args, rc)
    except Exception as e:
        rc = 1
        print(e, file=sys.stderr)
    return (rc, extra_info)

def main(args=None):
    """ Mimic streamtool using the REST api for ICP4D.
    """
    streamsx._streams._version._mismatch_check('streamsx.topology.context')

    rc, extra_info = run_cmd(args)
    return rc

def _parse_args(args):
    """ Argument parsing
    """
    cmd_parser = argparse.ArgumentParser(description='IBM Streams command line interface using REST.')
    cmd_parser.add_argument('--disable-ssl-verify', action='store_true', help='Disable SSL verification.')

    subparsers = cmd_parser.add_subparsers(help='Supported commands', dest='subcmd')

    _submitjob_parser(subparsers)
    _canceljob_parser(subparsers)
    _lsjobs_parser(subparsers)
    _lsappconfig_parser(subparsers)
    _rmappconfig_parser(subparsers)
    _mkappconfig_parser(subparsers)
    _chappconfig_parser(subparsers)
    _getappconfig_parser(subparsers)
    _rmtoolkit_parser(subparsers)
    _lstoolkit_parser(subparsers)
    _uploadtoolkit_parser(subparsers)
    _updateops_parser(subparsers)

    cmd_args = cmd_parser.parse_args(args)
    if cmd_args.subcmd is None:
        cmd_parser.print_help()
        exit(0)
    else:
        return cmd_args

def _user_arg(parser):
    parser.add_argument('--User', '-U', help='Specifies an IBM Streams user ID that has authority to run the command.', metavar='user')


if __name__ == '__main__':
    exit_code = main()
    sys.exit(exit_code)

